/**********************************************************/
/*                                                        */
/*     AutoMap / AutoLink 3.00 [beta 1]                   */
/*                                                        */
/*  version 1.1                                           */
/*  Eric Baland (AutoMap)                                 */
/*  & Koen Vrielink AutoLink)                             */
/*    January 1997                                        */
/*                                                        */
/*  Version 2.0                                           */
/*  Jasper Peeters (AutoLink)                             */
/*    July 1997                                           */
/*                                                        */
/*  Version 2.1                                           */
/*  Martial MICHEL (AutoMap & AutoLink :                  */
/*                    Concept & Programming)              */
/*  & Delphine GOUJON (AutoLink : Concept)                */
/*    July 1998                                           */
/*                                                        */
/*  Version 2.11                                          */
/*  Martial MICHEL                                        */
/*  & Thibaut ROUFFINEAU (Testing)                        */
/*    November 1998                                       */
/*                                                        */
/*  Version 2.11.1                                        */
/*  Martial MICHEL                                        */
/*    January 1999                                        */
/*                                                        */
/*  Version 2.11.2                                        */
/*  Martial MICHEL                                        */
/*    February 1999                                       */
/*                                                        */
/*  Version 2.20                                          */
/*  Martial MICHEL                                        */
/*    June 1999                                           */
/*                                                        */
/*  Version 3.00b1                                        */
/*  Martial MICHEL                                        */
/*  & Olivier NICOLAS (Testing & AutoMap : Concept)       */
/*    April 2000                                          */
/*                                                        */
/*  Version 3.00b2                                        */
/*  Martial MICHEL                                        */
/*  & Olivier NICOLAS (Stress Testing)                    */
/*    August 2000                                         */
/*                                                        */
/**********************************************************/

/************************************************************/
/*
  Contains : AutoLink (v3.00 [beta 2]) Common DEFINEs used
  Organization : NIST
  Developped for : use with AutoLink v3.00
  by  : Martial MICHEL (Guest Researcher at NIST / PhD student)
  Project Leader : Judith Ellen Devaney
  
  Comments or ideas are welcomed, please E-mail to :
  martial.michel@nist.gov
*/
/************************************************************/

#ifndef _AL_COMMON_H__
#define _AL_COMMON_H__

#if 0
static char _al_common_h_[] = 
"@(#)[v3.00 beta 2] AutoLink common DEFINEs used";
#endif

/************************************************************/
/*     mpi.h ... for MPI_Request                            */
/************************************************************/
#include <mpi.h>

/************************************************************/
/* User Modifiable values                                   */
/************************************************************/

/****************************************/
/*
  Hash Table size
  Note 1 : must be a prime number (for optimal distribution)
  Note 2 : part of the performance of Hash Table depends on it
           distribution of adresses will be done into this number slots
	   collision will be solved by search into the Single Linked List
  Note 3 : user can define one following the single previous rules 
           plus knowing that we choosed to use "int" as a type, so "limit"
	   to the value of the number should be 0< n <=32767
	   (hoping you will take a number at least > 1000 to make a basic
	   simple distribution)
*/
#ifndef AL_HASH_TABLE_PRIME_NUMBER
#define AL_HASH_TABLE_PRIME_NUMBER 10973
#endif

/****************************************/
/*
  Special Array Size declaration
  Note 1 : this size is only to determine the maximum number of elements
           contained in a SArray before creating an extension to this one
*/
#ifndef AL_SARRAY_SIZE
#define AL_SARRAY_SIZE 256
#endif

/****************************************/
/*
  Request List number of elements
  Note 1 : similar to the "Special Array Size declaration"
  Note 2 : Number of requests are dependant on the splitting inside Packet
           and number of different data-types
*/

#ifndef AL_REQLIST_SIZE
#define AL_REQLIST_SIZE 32
#endif

/****************************************/
/*
  "I" Request List number of elements
  Note 0 : Used for non blocking operations
  Note 1 : similar to the "Special Array Size declaration"
  Note 2 : Number of requests are dependant on the splitting inside Packet
           and number of different data-types
*/

#ifndef AL_IREQLIST_SIZE
#define AL_IREQLIST_SIZE 32
#endif

/****************************************/
/*
  Default Maximum Packet Size (in bytes)
*/
#ifndef AL_DEFAULTMAXPACKET
#define AL_DEFAULTMAXPACKET 4096
#endif

/****************************************/
/*
  Allow Users to change PacketSize dynamically
  Note 1 : if uncommented, consistency between processes is to be assured
           by users
  Note 2 : if commented, users can not use the AL_SetPacketSize functions
  Note 3 : to be mostly used by "Fine Packet Tuning" test case
*/

/* #define AL_USERCONTROL_PACKET */

/****************************************/
/*
  Define the number of Blocks per type created at a time
  (one block is one PACKET for each type created)
*/
#ifndef AL_BLOCKS_NUMBER
#define AL_BLOCKS_NUMBER 10
#endif

/****************************************/
/*
  Define the number of entries to create in advance for
  all pending IRecv entries
*/
#ifndef AL_IRECV_PENDING
#define AL_IRECV_PENDING 10
#endif

/****************************************/
/*
  Debug level
*/
/* Should you want to debug, 
   uncomment the following statement */
/* #define _AL_DEBUG_ */

#ifdef _AL_DEBUG_
/* Do not change the following values */
/* _AL_NODEBUG      : Basic level that should always be set */
/* _AL_DBG_FLUSH    : Flush output into the file ... may change perfs */
/* _AL_DBG_FNAME    : Print the function name entered */
/* _AL_DBG_FINT     : Tell the basic functions done by the function entered */
/* _AL_DBG_DATA     : Tell the status of the processed data */
/* _AL_DBG_STEP     : Follow the step of some "for" functions */
/* _AL_DBG_BENCH    : Will do Time benchmarking */
/* _AL_DBG_UBENCH   : Will add user text into log if set */
/* _AL_DBG_REQUEST  : To follow MPI_Request operations */
#define _AL_NODEBUG      0
#define _AL_DBG_FLUSH    1
#define _AL_DBG_FNAME    2
#define _AL_DBG_FINT     4
#define _AL_DBG_DATA     8
#define _AL_DBG_STEP     16
#define _AL_DBG_BENCH    32
#define _AL_DBG_UBENCH   64
#define _AL_DBG_REQUEST  128

/* End of "do not change" */

/* If debug is allowed, set debug level
   by making "|" (bitwise operator) between
   possible debug levels 
   Note that if debug is allowed, you need at least _AL_NODEBUG */
#ifndef _AL_DBG_LVL
#define _AL_DBG_LVL (_AL_NODEBUG | _AL_DBG_UBENCH)
#endif

/* Uncomment the next define, and set a (full path included) filename
   to store the log of your process in "filename.nodenumber" file */
/* #define _AL_DEBUG_FILENAME "ALlog" */

#endif /* #ifdef _AL_DEBUG_ */

/* Just to be on the safe side */
#ifdef _AL_DEBUG_
#ifndef _AL_DBG_LVL
#define _AL_DBG_LVL _AL_NODEBUG
#endif
#endif

/************************************************************/
/* End of User Modifiable values                            */
/************************************************************/

/************************************************************/
/* Internal specialized values                              */
/* Modify them can break the internal functions of AutoLink */
/************************************************************/

/****************************************/
/*
  NO CHILD pointer value 
  Note 1 : Can NOT be negative ("unsigned long int")
  Note 2 : Must be different for evey other possible value
*/
#define AL_NOCHILD (0xFFFFFFFF)

/****************************************/
/*
  True / False declarations
*/
#define AL_True   (1==1)
#define AL_False  (1==2)

/****************************************/
/*
  Error Return
*/
#define AL_Error (-1)

/************************************************************/
/* End of Internal specialized values                       */
/************************************************************/

#endif /* _AL_COMMON_H__ */

