/**********************************************************/
/*                                                        */
/*     AutoMap / AutoLink 3.00 [beta 2]                   */
/*                                                        */
/*  version 1.1                                           */
/*  Eric Baland (AutoMap)                                 */
/*  & Koen Vrielink (AutoLink)                            */
/*    January 1997                                        */
/*                                                        */
/*  Version 2.0                                           */
/*  Jasper Peeters (AutoLink)                             */
/*    July 1997                                           */
/*                                                        */
/*  Version 2.1                                           */
/*  Martial MICHEL (AutoMap & AutoLink :                  */
/*                    Concept & Programming)              */
/*  & Delphine GOUJON (AutoLink : Concept)                */
/*    July 1998                                           */
/*                                                        */
/*  Version 2.11                                          */
/*  Martial MICHEL                                        */
/*  & Thibaut ROUFFINEAU (Testing)                        */
/*    November 1998                                       */
/*                                                        */
/*  Version 2.11.1                                        */
/*  Martial MICHEL                                        */
/*    January 1999                                        */
/*                                                        */
/*  Version 2.11.2                                        */
/*  Martial MICHEL                                        */
/*    February 1999                                       */
/*                                                        */
/*  Version 2.20                                          */
/*  Martial MICHEL                                        */
/*    June 1999                                           */
/*                                                        */
/*  Version 3.00b1                                        */
/*  Martial MICHEL                                        */
/*  & Olivier NICOLAS (Testing & AutoMap : Concept)       */
/*    April 2000                                          */
/*                                                        */
/*  Version 3.00b2                                        */
/*  Martial MICHEL                                        */
/*  & Olivier NICOLAS (Stress Testing)                    */
/*    August 2000                                         */
/*                                                        */
/**********************************************************/

/************************************************************/
/*
  Contains : AutoLink (v3.00 [beta 2]) Internal data-types definitions
  Organization : NIST
  Developped for : use with AutoLink v3.00
  by  : Martial MICHEL (Guest Researcher at NIST / PhD student)
  Project Leader : Judith Ellen Devaney
  
  Comments or ideas are welcomed, please E-mail to :
  martial.michel@nist.gov
*/
/************************************************************/

#if 0
static char _al_internals_c_[] = 
"@(#)[v3.00 beta 2] AutoLink internal data-types definitions";
#endif

#include <stdlib.h>
#include <stdio.h>
#include "al_internals.h"

/**********************************************************/
/*     Version Mutual exclusion protocol                  */
/**********************************************************/

#include "autolink.h"

/* Version number check : number * 100 (need integer) */
#if ((! defined AMVERSION) || (! defined AMVERSION_BETA) || (AMVERSION != 300) || (AMVERSION_BETA != 2))
#error "ERROR : This version of AutoLink is designed to work only with entries generated from AutoMap v3.00 (beta 2)"
#endif

/************************************************************/
/*
  External values & functions
*/

extern int _AL_MaxRanks;

extern void (* _AL_GiveBlockBuffer[_AL_TpUsed])();

extern AL_PIRRLList      *AL_PIRRLList_freeList;
extern AL_PendingIRecvRL *AL_PendingIRecvRLHandler;

/************************************************************/
/*
  User hidden function
*/

void _AL_LogEntry(char *entry);

/************************************************************/
/*
  Single Linked List Definition
*/
/************************************************************/

AL_LinkedList*
AL_LL_Create()
{
  return (AL_LinkedList*) malloc(sizeof(AL_LinkedList));
}

void
AL_LL_Delete(AL_LinkedList* p)
{
  AL_LinkedList* tmp;
  
  while (p != NULL) {
    tmp = p->next;
    free(p);
    p = tmp;
  }
}

void
AL_LL_AddFirst(AL_LinkedList** pop, unsigned long int key, unsigned long int val)
{
  AL_LinkedList* tmp;
  
  tmp = AL_LL_Create();
  
  tmp->key = key;
  tmp->val = val;
  tmp->next = *pop;
  
  *pop = tmp;
}

unsigned long int
AL_LL_Search(AL_LinkedList* p, unsigned long int key)
{
  while (p != NULL)
    {
      if (p->key == key)
	return (unsigned long int) (p->val);
      p = p->next;
    }
  
  /* fail case */
  return (unsigned long int) (-1);
}


/************************************************************/
/*
  Hash Table using Single Linked List definition
*/
/************************************************************/


AL_HashTable*
AL_HT_Create()
{
  AL_HashTable* tmp;
  int i;

  tmp = malloc(sizeof(*tmp));

  for (i=0; i<AL_HASH_TABLE_PRIME_NUMBER; i++) {
    tmp->content[i] = NULL;
    tmp->number[i] = 0;
  }

  return (AL_HashTable*) tmp;
}

void
AL_HT_Delete(AL_HashTable** pop)
{
  AL_HT_Clear(*pop);

  free(*pop);
}

void
AL_HT_Clear(AL_HashTable* p)
{
  int i;
  
  for (i=0; i<AL_HASH_TABLE_PRIME_NUMBER; i++) {
    AL_LL_Delete(p->content[i]);
    p->number[i] = 0;
  }
}

void
AL_HT_Add(AL_HashTable* p, unsigned long int key, unsigned long int val)
{
  unsigned long int tmp = AL_HT_HFunction(key);

  AL_LL_AddFirst(&(p->content[tmp]), key, val);
  p->number[tmp]++;
}

unsigned long int
AL_HT_Search(AL_HashTable* p, unsigned long int key)
{
  return (unsigned long int) AL_LL_Search(p->content[AL_HT_HFunction(key)], key);
}

int
AL_HT_HFunction(unsigned long int adr)
{
  return (int)
    (
     (
      (adr & 0xffff00 ) 
      | ((adr & 0xff) << 24) 
      | ((adr & 0xff000000) >> 24)
      )
     % AL_HASH_TABLE_PRIME_NUMBER
     );
}

/************************************************************/
/*
  Special Array Definition
*/
/************************************************************/

AL_SArray*
AL_SArray_Create()
{
  AL_SArray* tmp;
  unsigned int i;

  tmp = malloc(sizeof(*tmp));
  tmp->first_available = 0;
  tmp->next = NULL;
  for ( i = 0; i < AL_SARRAY_SIZE; i++)
    tmp->content[i] = NULL;

  return (tmp);
}

void
AL_SArray_Delete(AL_SArray* p)
{ /* AL_SArray_Delete */
  /* 
     Note 1 : This function will only delete the Special Array itself
     not its content, for the Special Array is designed to
     be able to content anything at all, so it is recommanded
     to make sure of emptying the content before deleting the
     Special Array
  */

  AL_SArray *tmp, *tmp1, *tmp2, *tmp3;
  unsigned int i;

  tmp = p;
  while (tmp != NULL) {
    
    for (i = 0; i < tmp->first_available; i++) {
      tmp1 = tmp->content[i];
      
      while (tmp1 != NULL) {
	tmp2 = tmp1->next;
	free(tmp1);
	tmp1 = tmp2;
      }
    }
    
    tmp3 = tmp->next;
    free(tmp);
    tmp = tmp3;
  }
}

unsigned int
AL_SArray_GetMaxColumns(AL_SArray *p, unsigned int line)
{ /* GetMaxColumns */
  unsigned int to_go;
  AL_SArray *tmp;

  to_go = line;
  tmp = p;

  while (to_go > AL_SARRAY_SIZE-1) {
    /* Not in this part of the SArray */
    tmp = tmp->next;

    if (tmp == NULL)
      /* There is no such index */
      return (0);
    
    to_go -= AL_SARRAY_SIZE;
  }

  /* We are in the correct portion */
  if (tmp->content[to_go] == NULL)
    /* There is no element yet */
    return (0);
  else
    /* return the number of elements */
    return (_AL_SArray_Contains(tmp->content[to_go]));
}

unsigned int
_AL_SArray_Contains(AL_SArray *p)
{ /* _AL_SArray_Contains */
  AL_SArray* tmp = p;
  unsigned int result = 0;

  while (tmp->next != NULL) { 
    result += AL_SARRAY_SIZE;
    tmp = tmp->next;
  }

  result += tmp->first_available;

  return (result);
}

unsigned int
AL_SArray_AddLast(AL_SArray **pop, unsigned int line, void *value)
{ /* AL_SArray_AddLast */

  unsigned int to_go;
  unsigned int column = 0;
  AL_SArray **tmp;

  /* First, let's go to the correct line */
  to_go = line;
  tmp = pop;

  while (to_go > AL_SARRAY_SIZE-1) {
    /* Not in this part of the SArray */

    if ((*tmp)->next == NULL) {
      /* There is no such index, we need to create it */
      (*tmp)->next = AL_SArray_Create();
      /* And extend the number of elements of the current seen */
      (*tmp)->first_available = AL_SARRAY_SIZE;
    }
    
    /* Let's go to the next block */
    tmp = &((*tmp)->next);
    to_go -= AL_SARRAY_SIZE;
  }

  /* In this portion of the SArray */
  /* Let's adapt the index of the first available line if necessary */
  if (to_go >= (*tmp)->first_available)
    (*tmp)->first_available = to_go+1;
  
  tmp = (void*) &((*tmp)->content[to_go]);
  
  /* There is no content to the line, we need to create it */
  if ((*tmp) == NULL)
    (*tmp) = AL_SArray_Create();

  /* Now let's go and find the new available index */
  while ((*tmp)->next != NULL) {
    tmp = &((*tmp)->next);
    /* Update the return value to one more SArray crossed */
    column += AL_SARRAY_SIZE;
  }

  if ((*tmp)->first_available == AL_SARRAY_SIZE) {
    /* If there is no more available space in this set */
    /* let's add a new Special Array to it */
    (*tmp)->next = AL_SArray_Create();
    /* And we now need to go to the new available set */
    tmp = &((*tmp)->next);
    /* Let's then add a full SArray to the return value */
    column += AL_SARRAY_SIZE;
  }

  /* Now we can put the new value in the first available entry */
  (*tmp)->content[(*tmp)->first_available] = value;
  /* Extend the value of the return column */
  column += (*tmp)->first_available;
  /* And extend the position of the first available index */
  (*tmp)->first_available++;

  return (column);
}

void
AL_SArray_SetValue(AL_SArray **pop, unsigned int line, unsigned int col,
		   void *value)
{ /* AL_SArray_SetValue */
  AL_SArray **tmp;
  unsigned int to_go;

  /* First, let's go to the correct line */
  to_go = line;
  tmp = pop;
  
  while (to_go > AL_SARRAY_SIZE-1) {
    /* Not in this portion of the SArray */
    
    if  ((*tmp)->next == NULL) {
      /* There is no such index, we need to create it */
      (*tmp)->next = AL_SArray_Create();
      /* And extend the number of elements of the current seen */
      (*tmp)->first_available = AL_SARRAY_SIZE;
    }
    
    /* Let's go to the next block */
    tmp = &((*tmp)->next);
    to_go -= AL_SARRAY_SIZE;
  }
  
  /* In this portion of the SArray */
  
  /* Let's adapt the index of the first available line if necessary */
  if (to_go >= (*tmp)->first_available)
    (*tmp)->first_available = to_go+1;
  
  tmp = (void*) &((*tmp)->content[to_go]);

  /* Then let's get the exact column */
  to_go = col;
  
  if ((*tmp) == NULL)
    /* There is no content to the line, we need to create it */
    (*tmp) = AL_SArray_Create();
  
  while (to_go > AL_SARRAY_SIZE-1) {
    /* Not in this portion of the SArray */
    
    if ((*tmp)->next == NULL) { 
      /* There is no such index, we need to create it */
      (*tmp)->next = AL_SArray_Create();
      /* And extend the number of elements of the current seen */
      (*tmp)->first_available = AL_SARRAY_SIZE;
    }
    
    /* Let's go to the next block */
    tmp = &((*tmp)->next);
    to_go -= AL_SARRAY_SIZE;
  }

  /* In this portion of the SArray */
    
  /* Let's adapt the index of the first avaible column if necessary */
  if (to_go >= (*tmp)->first_available)
    (*tmp)->first_available = to_go+1;
  
  tmp = (void*) &((*tmp)->content[to_go]);
  
  /* Now that we are at the correct element, let's set the value */
  *tmp = (void*) value;
}

void *
AL_SArray_GetValue(AL_SArray *p, unsigned int line, unsigned int col)
{ /* AL_SArray_GetValue */
  unsigned int to_go;
  AL_SArray *tmp;

  to_go = line;
  tmp = p;

  while (to_go > AL_SARRAY_SIZE-1) {
    /* Not in this part of the SArray */
    tmp = tmp->next;

    if (tmp == NULL)
      /* There is no such index */
      return (void*) (AL_Error);
    
    to_go -= AL_SARRAY_SIZE;
  }

  /* Now, let's go to the correct index in the line */
  tmp = tmp->content[to_go];
  to_go = col;
  
  while (to_go > AL_SARRAY_SIZE-1) {
    /* Not in this part of the SArray */
    tmp = tmp->next;
    
    if (tmp == NULL)
      /* There is no such index */
      return (void*) (AL_Error);
    
    to_go -= AL_SARRAY_SIZE;
  }
  
  return (tmp->content[to_go]);
}

/************************************************************/
/*
  Queue Definition
*/
/************************************************************/

AL_subQueue*
AL_subQueue_Create()
{ /* AL_subQueue_Create */
  return (AL_subQueue*) malloc(sizeof(AL_subQueue));
}


void
AL_subQueue_Remove(AL_subQueue* p)
{ /* AL_subQueue_Remove */
  /* Will remove the first element of the subQueue leaving the */
  /* coherency problems to the calling function/procedure      */
  
  free(p);
}

AL_Queue*
AL_Queue_Create()
{ /* AL_Queue_Create */
  AL_Queue* tmp;

  tmp = malloc(sizeof(*tmp));

  tmp->first = NULL;
  tmp->last  = NULL;

  return (tmp);
}

void
AL_Queue_Delete(AL_Queue* p)
{ /* AL_Queue_Delete */
  AL_subQueue* tmp;
  AL_subQueue* current;

  if (p != NULL) {
    current = p->first;
    
    while (current != NULL) {
      tmp = current->next;
      AL_subQueue_Remove(current);
      current = tmp;
    }
    
    free(p);
  }
}

void
AL_Queue_Add(AL_Queue* p, int type_nbr, int pos)
{ /* AL_Queue_Add */
  AL_subQueue* tmp;

  tmp = AL_subQueue_Create();
  
  tmp->type_number = type_nbr;
  tmp->pos_in_data = pos;
  tmp->next = NULL;

  if (AL_Queue_IsEmpty(p) == AL_True) {
    p->first = tmp;
    p->last  = tmp;
  } else {
    p->last->next = tmp;
    p->last = tmp;
  }
}

void
AL_Queue_Remove(AL_Queue* p)
{ /* AL_Queue_Remove */
  AL_subQueue* tmp;

  if (AL_Queue_IsEmpty(p) == AL_False) {
    tmp = p->first->next;
    AL_subQueue_Remove(p->first);
    p->first = tmp;
  }
}

int
AL_Queue_GetDatatype(AL_Queue* p)
{ /* AL_Queue_GetDatatype */
  if (AL_Queue_IsEmpty(p) == AL_False) {
    return (p->first->type_number);
  }
  
  return (AL_Error);
}

int
AL_Queue_GetPosition(AL_Queue* p)
{ /* AL_Queue_GetPosition */
  if (AL_Queue_IsEmpty(p) == AL_False) {
    return (p->first->pos_in_data);
  }
  
  return (AL_Error);
}

int
AL_Queue_IsIn(AL_Queue* p, int type_nbr, int pos)
{ /* AL_Queue_IsIn */
  AL_subQueue* current = p->first;
  
  while (current != NULL) {
    if ((current->type_number == type_nbr) && (current->pos_in_data == pos))
      return (AL_True);
    
    current = current->next;
  }
  return (AL_False);
}

int
AL_Queue_IsEmpty(AL_Queue* p)
{ /* AL_Queue_IsEmpty */
  if (p->first == NULL)
    return (AL_True);
  
  return (AL_False);
}

/************************************************************/
/*
  Block Single Linked List definition
*/
/************************************************************/

AL_BlockLinkedList*
AL_BLL_Create()
{ /* AL_BLL_Create */
  return (AL_BlockLinkedList *) malloc(sizeof(AL_BlockLinkedList));
}

void
AL_BLL_Delete(AL_BlockLinkedList* p)
{ /* AL_BLL_Delete */
  AL_BlockLinkedList* tmp;

  while (p != NULL) {
    tmp = p->next;
    free(p->blockaddress);
    free(p);
    p = tmp;
  }
}

void
AL_BLL_SimplyDelete(AL_BlockLinkedList* p)
{ /* AL_BLL_Delete */
  /* Difference with AL_BLL_Delete is that we do not erase "p->blockaddress" */

  AL_BlockLinkedList* tmp;

  while (p != NULL) {
    tmp = p->next;
    free(p);
    p = tmp;
  }
}

void
AL_BLL_AddFirst(AL_BlockLinkedList** pop, void* value)
{ /* AL_BLL_Create */
  AL_BlockLinkedList *tmp;
  
  tmp = AL_BLL_Create();

  tmp->blockaddress = value;
  tmp->next = *pop;

  *pop = tmp;
}

void
AL_BLL_SwapFirst(AL_BlockLinkedList** pop1, AL_BlockLinkedList** pop2)
{ /* AL_BLL_SwapFirst */
  AL_BlockLinkedList *tmp = *pop1;

  *pop1 = (*pop1)->next;

  tmp->next = *pop2;

  *pop2 = tmp;
}

int
AL_BLL_Swap(AL_BlockLinkedList** pop1, AL_BlockLinkedList** pop2, void* adr)
{ /* AL_BLL_Swap */
  AL_BlockLinkedList *tmp1;
  AL_BlockLinkedList *tmp2 = *pop1;

  if (tmp2 == NULL)
    return (AL_False);

  if (tmp2->blockaddress == adr) {
    AL_BLL_SwapFirst(pop1, pop2);
    return(AL_True);
  }
  
  tmp1 = tmp2;
  tmp2 = tmp2->next;
  while (tmp2 != NULL) {
    if (tmp2->blockaddress == adr) {
      tmp1->next = tmp2->next;
      tmp2->next = *pop2;
      *pop2 = tmp2;
      
      return (AL_True);
    }

    tmp1 = tmp2;
    tmp2 = tmp2->next;
  }
      
  return (AL_False);
}

void *
AL_BLL_GetFirstValue(AL_BlockLinkedList* p)
{ /* AL_BLL_GetFirstValue */
  if (p == NULL)
    return (NULL);

  return (p->blockaddress);
}

int
AL_BLL_SearchValue(AL_BlockLinkedList* p, void* adr)
{ /* AL_BLL_SearchValue */
  AL_BlockLinkedList *tmp = p;

  while (tmp != NULL) {
    if (tmp->blockaddress == adr)
      return (AL_True);

    tmp = tmp->next;
  }

  return (AL_False);
}

/************************************************************/
/*
  Blocks definition
*/
/************************************************************/

AL_Blocks*
AL_Blocks_Create(int elementsize, int elements)
{ /* AL_Blocks_Create */
  AL_Blocks* tmp;
  int i;
  void* adrtmp;

  tmp = malloc(sizeof(*tmp));

  tmp->freeblocks = NULL;
  tmp->usedblocks = NULL;

  for (i = 0; i < elements; i++) {
    adrtmp = (void*) malloc(elementsize);
    AL_BLL_AddFirst( &(tmp->freeblocks), adrtmp);
  }

  tmp->incelements  = elements;
  tmp->elementsize  = elementsize;
  tmp->freeelements = elements;
  tmp->usedelements = 0;
  
  return (tmp);
}

void
AL_Blocks_Delete(AL_Blocks** pop)
{ /* AL_Blocks_Delete */

  /* Only free free blocks */
  AL_BLL_Delete((*pop)->freeblocks);
  /* Keeep "blocks" for used ones */
  AL_BLL_SimplyDelete((*pop)->usedblocks);

  free(*pop);
}

int
AL_Blocks_isFree(AL_Blocks* p)
{ /* AL_Blocks_isBlock */
  if (p->freeelements > 0)
    return (AL_True);

  return (AL_False);
}

void*
AL_Blocks_GetBlock(AL_Blocks** pop)
{ /* AL_Blocks_GetBlock */
  if ((*pop)->freeelements == 0)
    AL_Blocks_AddBlocks(pop, -1);

  AL_BLL_SwapFirst( &((*pop)->freeblocks), &((*pop)->usedblocks) );
  
  (*pop)->freeelements--;
  (*pop)->usedelements++;

  return ( (void *) AL_BLL_GetFirstValue( (*pop)->usedblocks ) );
}

void
AL_Blocks_GiveBlock(AL_Blocks** pop, void* adr)
{ /* AL_Blocks_GiveBlock */
  
  AL_BLL_Swap( &((*pop)->usedblocks), &((*pop)->freeblocks), adr);

  (*pop)->freeelements++;
  (*pop)->usedelements--;
}

int
AL_Blocks_isIn(AL_Blocks* p, void* adr)
{ /* AL_Blocks_isIn */
  if (AL_BLL_SearchValue(p->freeblocks, adr) == AL_True)
    return (AL_True);
  if (AL_BLL_SearchValue(p->usedblocks, adr) == AL_True)
    return (AL_True);

  return (AL_False);
}

int
AL_Blocks_isInUsed(AL_Blocks* p, void* adr)
{ /* AL_Blocks_isInUsed */
  return AL_BLL_SearchValue(p->usedblocks, adr);
}

int
AL_Blocks_isInFree(AL_Blocks* p, void* adr)
{ /* AL_Blocks_isInFree */
  return AL_BLL_SearchValue(p->freeblocks, adr);
}

void
AL_Blocks_AddBlocks(AL_Blocks** pop, int elements)
{ /* AL_Blocks_AddBlocks */
  int i;
  void* adrtmp;

  if (elements == -1)
    elements = (*pop)->incelements;

  for (i = 0; i < elements; i++) {
    adrtmp = (void*) malloc((*pop)->elementsize);
    AL_BLL_AddFirst( &((*pop)->freeblocks), adrtmp);
  }

  (*pop)->freeelements += elements;
}

/************************************************************/
/*
  Request List definition
*/
/************************************************************/

AL_ReqList*
AL_ReqList_Create()
{
  AL_ReqList* tmp;

  tmp = malloc(sizeof(*tmp));
  tmp->first_available = 0;
  tmp->next = NULL;

  return (tmp);
}

void
AL_ReqList_Delete(AL_ReqList* p)
{ /* AL_ReqList_Delete */
  /* 
     Note 1 : This function will delete the special array and 
     transfer used blocks back to free blocks
  */

  AL_Internals ALint;
  AL_ReqList *tmp, *tmp2;
  int i;

  tmp = p;
  while (tmp != NULL) {
    for (i = 0; i < tmp->first_available; i++) {
      ALint.AL_Data[tmp->type[i]] = (void *) tmp->content[i];
      _AL_GiveBlockBuffer[tmp->type[i]](&ALint);
    }
    
    tmp2 = tmp->next;
    free(tmp);
    tmp = tmp2;
  }
}

void
AL_ReqList_Add(AL_ReqList **pop, void *value, int type)
{ /* AL_ReqList_Add */
  AL_ReqList **tmp;

  tmp = pop;

  while ((*tmp)->next != NULL)
    /* Let's go to the next block */
    tmp = &((*tmp)->next);

  if ((*tmp)->first_available == AL_REQLIST_SIZE) {
    /* we need a new block */
    (*tmp)->next = AL_ReqList_Create();
    tmp = &((*tmp)->next);
  }
  
  (*tmp)->content[(*tmp)->first_available] = value;
  (*tmp)->type[(*tmp)->first_available] = type;
  (*tmp)->first_available++;
}

/************************************************************/
/*
  "I" Request List definition
*/
/************************************************************/

AL_IIRL*
AL_IIRL_Create()
{ /* AL_IIRL_Create */
  AL_IIRL* tmp;

  tmp = malloc(sizeof(*tmp));

  tmp->size           = AL_IREQLIST_SIZE;
  tmp->firstavailable = 0;
  tmp->datatype       = malloc(AL_IREQLIST_SIZE * sizeof(*tmp->datatype));
  tmp->request        = malloc(AL_IREQLIST_SIZE * sizeof(*tmp->request));
  tmp->status         = malloc(AL_IREQLIST_SIZE * sizeof(*tmp->status));
  tmp->block          = malloc(AL_IREQLIST_SIZE * sizeof(*tmp->block));

  return (tmp);
}

void
AL_IIRL_Destroy(AL_IIRL *p)
{ /* AL_IIRL_Destroy */
  int i;

  for (i = 0; i< p->firstavailable; i++)
    free(p->request[i]);

  free(p->datatype);
  free(p->request);
  free(p->status);
  free(p->block);

  free(p);
}

void
AL_IIRL_Realloc(AL_IIRL *p)
{ /* AL_IIRL_Realloc */
  p->size    += AL_IREQLIST_SIZE;
  p->datatype = realloc(p->datatype, p->size * sizeof(*p->datatype));
  p->request  = realloc(p->request, p->size * sizeof(*p->request));
  p->status   = realloc(p->status, p->size * sizeof(*p->status));
  p->block    = realloc(p->block, p->size * sizeof(*p->block));
} 

void
AL_IIRL_AddLast(AL_IIRL *p, int datatype, MPI_Request *request, void *block)
{ /* AL_IIRL_AddLast */
  if (p->firstavailable >= p->size)
    AL_IIRL_Realloc(p);

  p->datatype[p->firstavailable] = datatype;
  p->request[p->firstavailable]  = request;
  p->block[p->firstavailable]    = block;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_REQUEST)
  {
    char text[120];
    sprintf(text, "[AL_IIRL_AddLast] %d / state : %d",
	    p->firstavailable + 1, *(p->request[p->firstavailable]));
    _AL_LogEntry(text);
  }
#endif
#endif

  p->firstavailable++;
}

int
AL_IIRL_Waitall(AL_IIRL *p)
{ /* AL_IIRL_Waitall */
  /* 
     - For debugging capability, we will use a "for" here
     - Plus, we perform tests until we actually may have to wait.
     - It is to be noted that this comportment was added mostly for
     debugging some comportment in the misc MPI libraries available
     after we encountered some strange behaviors
  */
  int i;
  MPI_Status tmp;
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_REQUEST)
  char text[120];
#endif
#endif

  for (i = 0; i < p->firstavailable; i++) {

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_REQUEST)
    sprintf(text, "[AL_IIRL_Waitall] %d on %d / state : %d",
	    i + 1, p->firstavailable, *(p->request[i]));
    _AL_LogEntry(text);
#endif
#endif

    if (*(p->request[i]) != MPI_REQUEST_NULL) {
      int result;

      _AL_ReturnCode = MPI_Test(p->request[i], &result, &tmp);
      
      if (result == AL_True) {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_REQUEST)	
	sprintf(text, "[AL_IIRL_Waitall] (test) %d on %d / Changed state : %d",
		i + 1, p->firstavailable, *(p->request[i]));
	_AL_LogEntry(text);
#endif
#endif
      } else {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_REQUEST)
	sprintf(text, "[AL_IIRL_Waitall] (will wait) %d on %d / state : %d",
		i + 1, p->firstavailable, *(p->request[i]));
	_AL_LogEntry(text);
#endif
#endif

	_AL_ReturnCode = MPI_Wait(p->request[i], &tmp);
	
	if (_AL_ReturnCode == MPI_SUCCESS) {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_REQUEST)
	  sprintf(text, "[AL_IIRL_Waitall] %d on %d / Changed state : %d",
		  i + 1, p->firstavailable, *(p->request[i]));
	  _AL_LogEntry(text);
#endif
#endif
	} else {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_REQUEST)
	  sprintf(text, "[AL_IIRL_Waitall] %d / state : %d -- PB",
		  i + 1, *(p->request[i]));
	  _AL_LogEntry(text);
#endif
#endif
	  
	  return(AL_False);
	}
      }
    }
  }

  return (AL_True);
}

int
AL_IIRL_Testall(AL_IIRL *p)
{ /* AL_IIRL_Testall */
  /* 
     - For debugging capability, we will use a "for" here
     - It is to be noted that this comportment was added mostly for
     debugging some comportment in the misc MPI libraries available
     after we encountered some strange behaviors
  */
  int i, result;
  MPI_Status tmp;
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_REQUEST)
  char text[120];
#endif
#endif

  for (i = 0; i < p->firstavailable; i++) {

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_REQUEST)
    sprintf(text, "[AL_IIRL_Testall] %d on %d / state : %d",
	    i + 1, p->firstavailable, *(p->request[i]));
    _AL_LogEntry(text);
#endif
#endif

    if (*(p->request[i]) != MPI_REQUEST_NULL) {
      _AL_ReturnCode = MPI_Test(p->request[i], &result, &tmp);
     
      if (result == AL_True) {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_REQUEST)
    	sprintf(text, "[AL_IIRL_Testall] %d on %d / Changed state : %d",
		i + 1, p->firstavailable, *(p->request[i]));
	_AL_LogEntry(text);
#endif
#endif
      } else {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_REQUEST)
	sprintf(text, "[AL_IIRL_Testall] %d / state : %d -- NOT COMPLETED",
		i + 1, *(p->request[i]));
	_AL_LogEntry(text);
#endif
#endif

	return(AL_False);
      }
    }
  }
  
  return (AL_True);
}

void
AL_IIRL_GiveBlockBuffer(AL_IIRL* p, AL_Internals *ALint)
{ /* AL_IIRL_GiveBlockBuffer */
  int i;

  for (i = 0; i< p->firstavailable; i++) {
    /* Let's set the block to be deleted */
    ALint->AL_Data[p->datatype[i]] = p->block[i];

    _AL_GiveBlockBuffer[p->datatype[i]](ALint);
  }
}

/************************************************************/
/*
  "I" Send RL definitions
*/

AL_ISendRL*
AL_ISendRL_Create(void *buf, int datatype, int dest, int tag,
		  MPI_Comm comm, AL_Internals *ALint, AL_InitMsg *InitMsg)
{ /* AL_ISendRL_Create */
  AL_ISendRL* tmp;

  tmp = malloc(sizeof(*tmp));
  tmp->buf       = buf;
  tmp->datatype  = datatype;
  tmp->dest      = dest;
  tmp->tag       = tag;
  tmp->comm      = comm;
  tmp->InitMsg   = InitMsg;
  tmp->InitMsgRq = NULL;
  tmp->State     = AL_ISendRL_State_UNDEF;
  tmp->ALint     = ALint;

  tmp->IIRL      = AL_IIRL_Create();

  return (tmp);
}

void
AL_ISendRL_Delete(AL_ISendRL* p)
{ /* AL_ISendRL_Delete */
  AL_IIRL_Destroy(p->IIRL);
  
  free(p->InitMsg);
  free(p->InitMsgRq);

  free(p->ALint);

  free(p);
}

void
AL_ISendRL_SetInitMsgRq(AL_ISendRL* p, MPI_Request* InitMsgRq)
{ /* AL_ISendRL_SetInitMsg */
  p->InitMsgRq = InitMsgRq;
}

int
AL_ISendRL_WaitInitMsg(AL_ISendRL *p)
{ /* AL_ISendRL_WaitInitMsg */
  MPI_Status tmp;

  _AL_ReturnCode = MPI_Wait(p->InitMsgRq, &tmp);

  if (_AL_ReturnCode == MPI_SUCCESS)
    return (AL_True);

  return(AL_False);
}

int
AL_ISendRL_TestInitMsg(AL_ISendRL *p)
{ /* AL_ISendRL_TestInitMsg */
  MPI_Status tmp;
  int result;

  _AL_ReturnCode = MPI_Test(p->InitMsgRq, &result, &tmp);

  if (result == AL_True)
    return (AL_True);

  return(AL_False);
}

void
AL_ISendRL_GiveBlockBuffer(AL_ISendRL *p)
{ /* AL_ISendRL_GiveBlockBuffer */
  AL_IIRL_GiveBlockBuffer(p->IIRL, p->ALint);
}

/************************************************************/
/*
  "I" Recv RL definitions
*/

AL_IRecvRL*
AL_IRecvRL_Create(void **buf, int *datatype, int source, int tag,
		  MPI_Comm comm, AL_Internals *ALint, AL_ReqList **rqlist)
{ /* AL_IRecvRL_Create */
  AL_IRecvRL* tmp;
  
  tmp = malloc(sizeof(* tmp));

  tmp->buf       = buf;
  tmp->datatype  = datatype;
  tmp->source    = source;
  tmp->tag       = tag;
  tmp->comm      = comm;
  tmp->InitMsg   = malloc(sizeof(AL_InitMsg));
  tmp->InitMsgRq = NULL;
  tmp->State     = AL_IRecvRL_State_UNDEF;
  tmp->ALint     = ALint;
  tmp->rqlist     = rqlist;

  tmp->IIRL      = AL_IIRL_Create();

  if (comm == MPI_COMM_WORLD)
    tmp->realrank = source;
  else {
    MPI_Group tmp1, tmp2;
    int rank1, rank2;

    rank1 = source;

    MPI_Comm_group(comm, &tmp1);
    MPI_Comm_group(MPI_COMM_WORLD, &tmp2);

    MPI_Group_translate_ranks(tmp1, 1, &rank1, tmp2, &rank2);

    tmp->realrank = rank2;
  }

  return (tmp);
}

void
AL_IRecvRL_Delete(AL_IRecvRL* p)
{ /* AL_RecvReqList_Delete */
  AL_IIRL_Destroy(p->IIRL);

  free(p->InitMsg);
  free(p->InitMsgRq);

  free(p->ALint);
  
  free(p);
}

void
AL_IRecvRL_SetInitMsgRq(AL_IRecvRL* p, MPI_Request* InitMsgRq)
{ /* AL_IRecvRL_SetInitMsg */
  p->InitMsgRq = InitMsgRq;
}

int
AL_IRecvRL_WaitInitMsg(AL_IRecvRL *p)
{ /* AL_IRecvRL_WaitInitMsg */
  MPI_Status tmp;

  _AL_ReturnCode = MPI_Wait(p->InitMsgRq, &tmp);

  if (_AL_ReturnCode == MPI_SUCCESS)
    return (AL_True);
  
  return(AL_False);
}

int
AL_IRecvRL_TestInitMsg(AL_IRecvRL *p)
{ /* AL_IRecvRL_TestInitMsg */
  MPI_Status tmp;
  int result;

  _AL_ReturnCode = MPI_Test(p->InitMsgRq, &result, &tmp);

  if (result == AL_True)
    return (AL_True);

  return(AL_False);
}

void
AL_IRecvRL_GiveBlockBuffer(AL_IRecvRL *p)
{ /* AL_IRecvRL_GiveBlockBuffer */

  /* We actually only give them back _only_ if we the user is not using them */
  if (p->rqlist == NULL)
    AL_IIRL_GiveBlockBuffer(p->IIRL, p->ALint);
}

/****************************************/
/*
  Pending I Recv Requests definitions
*/

void
AL_PIRRLList_AddFreeElements()
{ /* AL_PIRRLList_AddFreeElements */
  int i;

  for (i = 0; i < AL_IRECV_PENDING; i++)
    AL_PIRRLList_AddHeadNew(&AL_PIRRLList_freeList, NULL);
}

void
AL_PIRRLList_DeleteFreeElemenst()
{ /* AL_PIRRLList_DeleteFreeElements */
  AL_PIRRLList *tmp = AL_PIRRLList_freeList;
  AL_PIRRLList *tmp2;

  while (tmp != NULL) {
    tmp2 = tmp->next;
    free(tmp);
    tmp = tmp2;
  }
}

AL_PIRRLList*
AL_PIRRLList_GetHeadFreeElements()
{ /* AL_PIRRLList_GetHeadFreeElements */
  if (AL_PIRRLList_freeList == NULL)
    AL_PIRRLList_AddFreeElements();

  return (AL_PIRRLList_GetHead(&AL_PIRRLList_freeList));
}

void
AL_PIRRLList_GiveBackFreeElements(AL_PIRRLList **pop, AL_IRecvRL *rlist)
{ /* AL_PIRRLList_GiveBackFreeElements */
  AL_PIRRLList *tmp, *tmp2;

  /* Case 1 : First element */
  if ((*pop)->rlist == rlist) {
    tmp2 = *pop;
    tmp = tmp2->next;
    *pop = tmp;
    AL_PIRRLList_AddHead(&AL_PIRRLList_freeList, tmp2);
  } else {
    /* Case 2 : Otherwise */
    tmp2 = *pop;
    tmp = tmp2->next;;
    while (tmp != NULL) {
      if (tmp->rlist == rlist) {
	tmp2->next = tmp->next;
	AL_PIRRLList_AddHead(&AL_PIRRLList_freeList, tmp);
      }
      
      tmp2 = tmp;
      tmp = tmp2->next;
    }
  }
}

/**********/

AL_PIRRLList*
AL_PIRRLList_Create(AL_IRecvRL *rlist, AL_PIRRLList* next)
{ /* AL_PIRRLList_Create */
  AL_PIRRLList* tmp;

  tmp = malloc(sizeof(* tmp));

  tmp->rlist = rlist;
  tmp->next  = next;

  return (tmp);
}

void
AL_PIRRLList_Delete(AL_PIRRLList* p)
{ /* AL_PIRRLList_Delete */
  /*
    Note 1 : We do not take care of removing pending rlists, just all PIRRLList
  */

  AL_PIRRLList *tmp, *tmp2;

  tmp = p;
  while (tmp != NULL) {
    tmp2 = tmp->next;
    free(tmp);
    tmp = tmp2;
  }
}

void
AL_PIRRLList_AddHeadNew(AL_PIRRLList **pop, AL_IRecvRL *rlist)
{ /* AL_PIRRLList_AddHeadNew */
  /*
    Note 1 : This function is only called when adding new elements 
             to the freeList
  */

  AL_PIRRLList *tmp;

  tmp = AL_PIRRLList_Create(rlist, *pop);

  *pop = tmp;
}

void
AL_PIRRLList_SetRlist(AL_PIRRLList *p, AL_IRecvRL *rlist)
{ /* AL_PIRRLList_SetRlist */
  p->rlist = rlist;
}

void
AL_PIRRLList_AddHead(AL_PIRRLList **pop, AL_PIRRLList *p)
{ /* AL_PIRRLList_AddHead */
  p->next = *pop;
  *pop = p;
}

void
AL_PIRRLList_AddTail(AL_PIRRLList **pop, AL_IRecvRL *rlist, AL_PIRRLList *p)
{ /* AL_PIRRLList_AddTail */
  AL_PIRRLList *tmp, *tmp2;

  /* Case 1 : First element */
  if (*pop == NULL) {
    p->rlist = rlist;
    p->next  = *pop;

    *pop = p;
  } else {
    /* Case 2 : Otherwise */
    tmp = *pop;
    while (tmp != NULL) {
      tmp2 = tmp;
      tmp = tmp->next;
    }
    p->rlist = rlist;
    p->next  = tmp;
    tmp2->next = p;
  }
}

AL_PIRRLList*
AL_PIRRLList_GetHead(AL_PIRRLList **pop)
{ /* AL_PIRRLList_GetHead */
  AL_PIRRLList *tmp;

  tmp = *pop;

  if (tmp != NULL) {
    *pop = tmp->next;
    tmp->next = NULL;

    return (tmp);
  }
    
  return (NULL);
}

AL_IRecvRL*
AL_PIRRLList_GetPending(AL_PIRRLList **pop)
{ /* AL_PIRRLList_GetPending */
  /*
    Note 1 : An entry stays pending as long as the previous one is not out
    of the AL_IRecvRL_State_WAITIM stage so we return all entries equal
    or before that
  */

  AL_PIRRLList *tmp;

  tmp = *pop;
  while (tmp != NULL) {
    if ((tmp->rlist->State) <= _AL_STILLPENDINGSTATE)
      return (tmp->rlist);

    tmp = tmp->next;
  }

  return (NULL);
}

/********************/

AL_PIRRLTagList*
AL_PIRRLTagList_Create(int tag)
{ /* AL_PIRRLTagList_Create */
  /*
    Note 1 : All access/creation functions should position element in
    first position in the linked list
  */

  AL_PIRRLTagList *tmp;

  tmp = malloc(sizeof(*tmp));

  tmp->tag     = tag;
  tmp->entries = NULL;
  tmp->next    = NULL;

  return(tmp);
}

void
AL_PIRRLTagList_CreateNewTag(AL_PIRRLTagList **pop, int tag)
{ /* AL_PIRRLTagList_CreateNewTag */
  AL_PIRRLTagList *tmp;

  tmp = AL_PIRRLTagList_Create(tag);
  
  tmp->next = *pop;

  *pop = tmp;
}

void
AL_PIRRLTagList_Delete(AL_PIRRLTagList* p)
{ /* AL_PIRRLTagList_Delete */
  /*
    Note 1 : we do not delete entries
  */
  
  AL_PIRRLTagList *tmp, *tmp2;

  tmp = p;
  while (tmp != NULL) {
    tmp2 = tmp->next;
    free(tmp);
    tmp = tmp2;
  }
}

int
AL_PIRRLTagList_IsIn(AL_PIRRLTagList** pop, int tag)
{ /* AL_PIRRLTagList_IsIn */
  /*
    Note 1 : Put entry in front if in
  */
  AL_PIRRLTagList *tmp, *tmp2;

  /* Case 0 : No element */
  if (*pop == NULL)
    return(AL_False);

  tmp = *pop;
  /* Case 1 : First element */
  if (tmp->tag == tag)
    return(AL_True);
  
  tmp2 = tmp;
  tmp = tmp->next;
  /* Case 2 : Somewhere */
  while (tmp != NULL) {
    if (tmp->tag == tag) {
      tmp2->next = tmp->next;
      tmp->next  = (*pop)->next;
      *pop = tmp;

      return(AL_True);
    }
    tmp2 = tmp;
    tmp = tmp->next;
  }
  
  /* Case 3 : Nowhere */
  return(AL_False);
}

void
AL_PIRRLTagList_AddRlist(AL_PIRRLTagList **pop, AL_IRecvRL *rlist)
{ /* AL_PIRRLTagList_AddRlist */
  AL_PIRRLList *tmp;

  tmp = AL_PIRRLList_GetHeadFreeElements();

  AL_PIRRLList_AddTail(&((*pop)->entries), rlist, tmp);
}

AL_IRecvRL*
AL_PIRRLTagList_GetPending(AL_PIRRLTagList **pop)
{ /* AL_PIRRLTagList_GetPending */
  return AL_PIRRLList_GetPending(&((*pop)->entries));
}

void
AL_PIRRLTagList_GiveBack(AL_PIRRLTagList **pop, AL_IRecvRL *rlist)
{ /* AL_PIRRLTagList_GiveBack */
  AL_PIRRLList_GiveBackFreeElements(&((*pop)->entries), rlist);
}

/********************/ 
  
AL_PendingIRecvRL*
AL_PendingIRecvRL_Create()
{ /* AL_PendingIRecvRL_Create */
  int i;
  AL_PendingIRecvRL* tmp;

  tmp = malloc(sizeof(*tmp));

  tmp->pendingList = malloc(sizeof(AL_PIRRLTagList*) * _AL_MaxRanks);
  for (i = 0; i < _AL_MaxRanks; i++)
    tmp->pendingList[i] = NULL;
  
  return (tmp);
}

void
AL_PendingIRecvRL_Delete(AL_PendingIRecvRL *p)
{ /* AL_PendingIRecvRL_Delete */
  /*
    Note 1 : we delete all entries here (there should be none active left)
  */
  int i;

  for (i = 0; i < _AL_MaxRanks; i++)
    AL_PIRRLTagList_Delete(p->pendingList[i]);

  free(p->pendingList);
  free(p);
}

void
AL_PendingIRecvRL_AddRlist(AL_PendingIRecvRL **pop, AL_IRecvRL *rlist)
{ /* AL_PendingIRecvRL_AddRlist */
  AL_PIRRLTagList **tmp;

  tmp = &(((*pop)->pendingList)[rlist->realrank]);

  if (AL_PIRRLTagList_IsIn(tmp, rlist->tag) == AL_True) {
    AL_PIRRLTagList_AddRlist(tmp, rlist);
  } else {
    if (*tmp == NULL)
      *tmp = AL_PIRRLTagList_Create(rlist->tag);
    else
      AL_PIRRLTagList_CreateNewTag(tmp, rlist->tag);

    AL_PIRRLTagList_AddRlist(tmp, rlist);
  }
}

AL_IRecvRL*
AL_PendingIRecvRL_GetPending(AL_PendingIRecvRL **pop, int rank, int tag)
{ /* AL_PendingIRecvRL_GetPending */
  AL_PIRRLTagList **tmp;

  tmp = &(((*pop)->pendingList)[rank]);

  if (AL_PIRRLTagList_IsIn(tmp, tag) == AL_True)
    return AL_PIRRLTagList_GetPending(tmp);
  else
    return (NULL);
}

void
AL_PendingIRecvRL_GiveBack(AL_PendingIRecvRL **pop, AL_IRecvRL *rlist)
{ /* AL_PendingIRecvRL_AddRlist */
  AL_PIRRLTagList **tmp;

  tmp = &(((*pop)->pendingList)[rlist->realrank]);

  if (AL_PIRRLTagList_IsIn(tmp, rlist->tag) == AL_True)
    AL_PIRRLTagList_GiveBack(tmp, rlist);
}

/************************************************************/
/*
  End of "al_internals.c"
*/
/************************************************************/
