/**********************************************************/
/*                                                        */
/*     AutoMap / AutoLink 3.00 [beta 2]                   */
/*                                                        */
/*  version 1.1                                           */
/*  Eric Baland (AutoMap)                                 */
/*  & Koen Vrielink (AutoLink)                            */
/*    January 1997                                        */
/*                                                        */
/*  Version 2.0                                           */
/*  Jasper Peeters (AutoLink)                             */
/*    July 1997                                           */
/*                                                        */
/*  Version 2.1                                           */
/*  Martial MICHEL (AutoMap & AutoLink :                  */
/*                    Concept & Programming)              */
/*  & Delphine GOUJON (AutoLink : Concept)                */
/*    July 1998                                           */
/*                                                        */
/*  Version 2.11                                          */
/*  Martial MICHEL                                        */
/*  & Thibaut ROUFFINEAU (Testing)                        */
/*    November 1998                                       */
/*                                                        */
/*  Version 2.11.1                                        */
/*  Martial MICHEL                                        */
/*    January 1999                                        */
/*                                                        */
/*  Version 2.11.2                                        */
/*  Martial MICHEL                                        */
/*    February 1999                                       */
/*                                                        */
/*  Version 2.20                                          */
/*  Martial MICHEL                                        */
/*    June 1999                                           */
/*                                                        */
/*  Version 3.00b1                                        */
/*  Martial MICHEL                                        */
/*  & Olivier NICOLAS (Testing & AutoMap : Concept)       */
/*    April 2000                                          */
/*                                                        */
/*  Version 3.00b2                                        */
/*  Martial MICHEL                                        */
/*  & Olivier NICOLAS (Stress Testing)                    */
/*    August 2000                                         */
/*                                                        */
/**********************************************************/

/************************************************************/
/*
  Contains : AutoLink (v3.00 [beta 2]) Internal data-types definitions
  Organization : NIST
  Developped for : use with AutoLink v3.00
  by  : Martial MICHEL (Guest Researcher at NIST / PhD student)
  Project Leader : Judith Ellen Devaney
  
  Comments or ideas are welcomed, please E-mail to :
  martial.michel@nist.gov
*/
/************************************************************/

#ifndef _AL_INTERNALS_H__
#define _AL_INTERNALS_H__

#if 0
static char _al_internals_h_[] = 
"@(#)[v3.00 beta 2] AutoLink internal data-types definitions";
#endif

#include "al_common.h"
#include <mpi.h>
#include "mpitypes.h"

/****************************************/
/*
  AutoLink Internal MPI defined datatype
 */

typedef struct _AL_InitMsg 
{
  unsigned long int  DataCount[_AL_TpUsed];
  int                InitialDatatype;
} AL_InitMsg;

extern MPI_Datatype  MPI_AL_InitMsg;

/****************************************/
/*
  Single Linked List Structure
*/

typedef struct _AL_LinkedList AL_LinkedList;

struct _AL_LinkedList
{
  unsigned long int    key; /* Key for the HASH TABLE             */
  unsigned long int    val; /* Value for the "tag" of the element */
  AL_LinkedList *next;      /* Next in the Linked List            */
};


/****************************************/
/*
  Single Linked List Functions
*/

AL_LinkedList*
AL_LL_Create();

void
AL_LL_Delete(AL_LinkedList* p);

void
AL_LL_AddFirst(AL_LinkedList** pop, unsigned long int key,
	       unsigned long int val);

unsigned long int
AL_LL_Search(AL_LinkedList* p, unsigned long int key);

/****************************************/
/*
  Hash Table Structure
*/

typedef struct _AL_HashTable AL_HashTable;

struct _AL_HashTable
{
  AL_LinkedList*  content[AL_HASH_TABLE_PRIME_NUMBER]; 
  int number[AL_HASH_TABLE_PRIME_NUMBER];
};


/****************************************/
/*
  Hash Table Functions
*/

AL_HashTable*
AL_HT_Create();

void
AL_HT_Delete(AL_HashTable** pop);

void
AL_HT_Clear(AL_HashTable* p);

void
AL_HT_Add(AL_HashTable* p, unsigned long int key, unsigned long int val);

unsigned long int
AL_HT_Search(AL_HashTable* p, unsigned long int key);

int
AL_HT_HFunction(unsigned long int adr);

/****************************************/
/*
  Special Array Structure
*/

typedef struct _AL_SArray AL_SArray;

struct _AL_SArray
{
  unsigned int first_available; /* first available index */
  void* content[AL_SARRAY_SIZE];
  AL_SArray* next;
};

/****************************************/
/*
  Special Array Functions
*/

AL_SArray*
AL_SArray_Create();

void
AL_SArray_Delete(AL_SArray* p);

unsigned int
AL_SArray_GetMaxColumns(AL_SArray *p, unsigned int line);

unsigned int
_AL_SArray_Contains(AL_SArray *p);

unsigned int
AL_SArray_AddLast(AL_SArray **pop, unsigned int line, void *value);

void
AL_SArray_SetValue(AL_SArray **pop, unsigned int line, unsigned int col,
		   void *value);

void *
AL_SArray_GetValue(AL_SArray *p, unsigned int line, unsigned int col);

/****************************************/
/*
  Queue Structure
*/

typedef struct _AL_subQueue AL_subQueue;

struct _AL_subQueue
{
  int type_number;
  int pos_in_data;
  AL_subQueue*  next;
};

typedef struct _AL_Queue AL_Queue;

struct _AL_Queue
{
  AL_subQueue*   first;
  AL_subQueue*   last;
};

/****************************************/
/*
  Queue Functions
*/

AL_subQueue*
AL_subQueue_Create();

AL_Queue*
AL_Queue_Create();

void
AL_Queue_Delete(AL_Queue* p);

void
AL_Queue_Add(AL_Queue* p, int type_nbr, int pos);

void
AL_Queue_Remove(AL_Queue* p);

int
AL_Queue_GetDatatype(AL_Queue* p);

int
AL_Queue_GetPosition(AL_Queue* p);

int
AL_Queue_IsIn(AL_Queue* p, int type_nbr, int pos);

int
AL_Queue_IsEmpty(AL_Queue* p);

/****************************************/
/*
  Block Single Linked List Structure
*/

typedef struct _AL_BlockLinkedList AL_BlockLinkedList;

struct _AL_BlockLinkedList
{
  void                *blockaddress; /* Address of the block */
  AL_BlockLinkedList  *next;         /* Next BLL */
};

/****************************************/
/*
  Block Single Linked List Functions
*/

AL_BlockLinkedList*
AL_BLL_Create();

void
AL_BLL_Delete(AL_BlockLinkedList* p);

void
AL_BLL_SimplyDelete(AL_BlockLinkedList* p);

void
AL_BLL_AddFirst(AL_BlockLinkedList** pop, void* value);

void
AL_BLL_SwapFirst(AL_BlockLinkedList** pop1, AL_BlockLinkedList** pop2);

int
AL_BLL_Swap(AL_BlockLinkedList** pop1, AL_BlockLinkedList** pop2, void* adr);

void*
AL_BLL_GetFirstValue(AL_BlockLinkedList* p);

int
AL_BLL_SearchValue(AL_BlockLinkedList* p, void* adr);

/****************************************/
/*
  Blocks Structure
*/

typedef struct _AL_Blocks AL_Blocks;

struct _AL_Blocks
{
  AL_BlockLinkedList  *freeblocks;  /* Free Blocks Linked List */
  AL_BlockLinkedList  *usedblocks;  /* Used Blocks Linked List */
  unsigned int        incelements;  /* default number to inc is needed */
  unsigned int        elementsize;  /* Size of one element */
  unsigned int        freeelements; /* Number of free elements */
  unsigned int        usedelements; /* Number of used elements */
};

/****************************************/
/*
  Blocks Functions
*/

AL_Blocks*
AL_Blocks_Create(int elementsize, int elements);

void
AL_Blocks_Delete(AL_Blocks** pop);

int
AL_Blocks_isFree(AL_Blocks* p);

void*
AL_Blocks_GetBlock(AL_Blocks** pop);

void
AL_Blocks_GiveBlock(AL_Blocks** pop, void* adr);

int
AL_Blocks_isIn(AL_Blocks* p, void* adr);

int
AL_Blocks_isInUsed(AL_Blocks* p, void* adr);

int
AL_Blocks_isInFree(AL_Blocks* p, void* adr);

void
AL_Blocks_AddBlocks(AL_Blocks** pop, int elements);

/****************************************/
/*
  Request List Structure
*/

typedef struct _AL_ReqList AL_ReqList;

struct _AL_ReqList
{
  int first_available; /* first available index */
  int   type[AL_REQLIST_SIZE];
  void* content[AL_REQLIST_SIZE];
  AL_ReqList* next;
};

/****************************************/
/*
  ReqList Functions
*/

AL_ReqList*
AL_ReqList_Create();

void
AL_ReqList_Delete(AL_ReqList* p);

void
AL_ReqList_Add(AL_ReqList **pop, void *value, int type);

/****************************************/
/*
  AutoLink Internals
*/

typedef struct _AL_Internals {
  unsigned long int  AL_DataCount[_AL_TpUsed];
  void*              AL_Data[_AL_TpUsed];
  AL_Queue*          AL_Next;
  AL_HashTable*      AL_Mark;
  AL_SArray*         AL_Address;
} AL_Internals;

/****************************************/
/*
  "I" Request List Structure
*/

typedef struct _AL_IInternalReqList AL_IIRL;

struct _AL_IInternalReqList {
  int           size;
  int           firstavailable;
  int          *datatype;
  MPI_Request **request;
  MPI_Status   *status;
  void        **block;
};

/****************************************/
/*
  "I" ReqList Functions
*/

AL_IIRL*
AL_IIRL_Create();

void
AL_IIRL_Destroy(AL_IIRL *p);

void
AL_IIRL_Realloc(AL_IIRL *p);

void
AL_IIRL_AddLast(AL_IIRL *p, int datatype, MPI_Request *request, void *block);

int
AL_IIRL_Waitall(AL_IIRL *p);

int
AL_IIRL_Testall(AL_IIRL *p);

void
AL_IIRL_GiveBlockBuffer(AL_IIRL* p, AL_Internals *ALint);

/****************************************/
/*
  AutoLink Code return variable
  (needed since we are doing some MPI dependent code now in "al_internals.c")
*/
extern int _AL_ReturnCode;

/****************************************/
/*
  "I" Send ReqList Structure
*/

typedef struct _AL_ISendReqList AL_ISendRL;

struct _AL_ISendReqList
{
  void           *buf;
  int             datatype;
  int             dest;
  int             tag;
  MPI_Comm        comm;
  AL_InitMsg     *InitMsg;
  MPI_Request    *InitMsgRq;
  int             State;
  AL_Internals   *ALint;
  AL_IIRL        *IIRL;
};

/****************************************/
/*
  "I" Send ReqList State values
*/
enum {
  AL_ISendRL_State_UNDEF,  /* Default state       */
  AL_ISendRL_State_WAITIM, /* Waiting for InitMsg */
  AL_ISendRL_State_WAITMSG,/* Waiting for all Msg */
  AL_ISendRL_State_DONE    /* AllMsg Sent         */
};

/****************************************/
/*
  "I" Send ReqList functions
*/

AL_ISendRL*
AL_ISendRL_Create(void *buf, int datatype, int dest, int tag,
		  MPI_Comm comm, AL_Internals *ALint, AL_InitMsg *InitMsg);

void
AL_ISendRL_Delete(AL_ISendRL* p);

void
AL_ISendRL_SetInitMsgRq(AL_ISendRL* p, MPI_Request* InitMsgRq);

int
AL_ISendRL_WaitInitMsg(AL_ISendRL *p);

int
AL_ISendRL_TestInitMsg(AL_ISendRL *p);

void
AL_ISendRL_GiveBlockBuffer(AL_ISendRL *p);

/****************************************/
/*
  "I" Recv ReqList Structure
*/

typedef struct _AL_IRecvReqList AL_IRecvRL;

struct _AL_IRecvReqList
{
  void          **buf;
  int            *datatype;
  int             source;
  int             tag;
  MPI_Comm        comm;
  AL_InitMsg     *InitMsg;
  MPI_Request    *InitMsgRq;
  int             State;
  AL_Internals   *ALint;
  AL_IIRL        *IIRL;
  AL_ReqList    **rqlist;
  int             realrank;
};

/****************************************/
/*
  "I" Recv Req List State values
*/

enum {
  AL_IRecvRL_State_UNDEF,     /* Default state               */
  AL_IRecvRL_State_PENDING,   /* Can not wait on InitMsg yet */
  AL_IRecvRL_State_WAITIM,    /* Waiting for InitMsg         */
  AL_IRecvRL_State_WAITMSG,   /* Waiting for all Msg         */
  AL_IRecvRL_State_WAITMEM,   /* Recreating elements         */
  AL_IRecvRL_State_WAITLINKS, /* Waiting on links            */
  AL_IRecvRL_State_DONE       /* All links recreated         */
};


/****************************************/
/*
  "I" Recv ReqList functions
*/

AL_IRecvRL*
AL_IRecvRL_Create(void **buf, int *datatype, int source, int tag,
		  MPI_Comm comm, AL_Internals *ALint, AL_ReqList **rqlist);

void
AL_IRecvRL_Delete(AL_IRecvRL* p);

void
AL_IRecvRL_SetInitMsgRq(AL_IRecvRL* p, MPI_Request* InitMsgRq);

int
AL_IRecvRL_WaitInitMsg(AL_IRecvRL *p);

int
AL_IRecvRL_TestInitMsg(AL_IRecvRL *p);

void
AL_IRecvRL_GiveBlockBuffer(AL_IRecvRL *p);


/****************************************/
/*
  Pending I Recv Requests structures
*/

typedef struct _AL_PIRRLList AL_PIRRLList;

struct _AL_PIRRLList 
{
  AL_IRecvRL    *rlist;
  AL_PIRRLList  *next;
};

/********************/

typedef struct _AL_PIRRLTagList AL_PIRRLTagList;

struct _AL_PIRRLTagList
{
  int             tag;
  AL_PIRRLList    *entries;
  AL_PIRRLTagList *next;
};

/********************/

typedef struct _AL_PendingIRecvRL AL_PendingIRecvRL;

struct _AL_PendingIRecvRL
{
  AL_PIRRLTagList **pendingList;
};

/****************************************/
/*
  Pending I Recv Requests functions
*/

void
AL_PIRRLList_AddFreeElements();

void
AL_PIRRLList_DeleteFreeElements();

AL_PIRRLList*
AL_PIRRLList_GetHeadFreeElements();

void
AL_PIRRLList_GiveBackFreeElements(AL_PIRRLList **pop, AL_IRecvRL *rlist);

/**********/

AL_PIRRLList*
AL_PIRRLList_Create(AL_IRecvRL *rlist, AL_PIRRLList* next);

void
AL_PIRRLList_Delete(AL_PIRRLList* p);

void
AL_PIRRLList_AddHeadNew(AL_PIRRLList **pop, AL_IRecvRL *rlist);

void
AL_PIRRLList_SetRlist(AL_PIRRLList *p, AL_IRecvRL *rlist);

void
AL_PIRRLList_AddHead(AL_PIRRLList **pop, AL_PIRRLList *p);

void
AL_PIRRLList_AddTail(AL_PIRRLList **pop, AL_IRecvRL *rlist, AL_PIRRLList *p);

AL_PIRRLList*
AL_PIRRLList_GetHead(AL_PIRRLList **pop);

AL_IRecvRL*
AL_PIRRLList_GetPending(AL_PIRRLList **pop);

/********************/

AL_PIRRLTagList*
AL_PIRRLTagList_Create(int tag);

void
AL_PIRRLTagList_CreateNewTag(AL_PIRRLTagList **pop, int tag);

void
AL_PIRRLTagList_Delete(AL_PIRRLTagList *p);

int
AL_PIRRLTagList_IsIn(AL_PIRRLTagList **pop, int tag);

void
AL_PIRRLTagList_AddRlist(AL_PIRRLTagList **pop, AL_IRecvRL *rlist); 

AL_IRecvRL*
AL_PIRRLTagList_GetPending(AL_PIRRLTagList **pop);

void
AL_PIRRLTagList_GiveBack(AL_PIRRLTagList **pop, AL_IRecvRL *rlist);

/********************/

AL_PendingIRecvRL*
AL_PendingIRecvRL_Create();

void
AL_PendingIRecvRL_Delete(AL_PendingIRecvRL *p);

void
AL_PendingIRecvRL_AddRlist(AL_PendingIRecvRL **pop, AL_IRecvRL *rlist);

AL_IRecvRL*
AL_PendingIRecvRL_GetPending(AL_PendingIRecvRL **pop, int rank, int tag);

void
AL_PendingIRecvRL_GiveBack(AL_PendingIRecvRL **pop, AL_IRecvRL *rlist);

/****************************************/
/*
  Pending I Recv Requests functions
*/

#define _AL_STILLPENDINGSTATE AL_IRecvRL_State_WAITIM

#endif /* _AL_INTERNALS__H__ */
