/**********************************************************/
/*                                                        */
/*     AutoMap / AutoLink 3.00 [beta 2]                   */
/*                                                        */
/*  version 1.1                                           */
/*  Eric Baland (AutoMap)                                 */
/*  & Koen Vrielink (AutoLink)                            */
/*    January 1997                                        */
/*                                                        */
/*  Version 2.0                                           */
/*  Jasper Peeters (AutoLink)                             */
/*    July 1997                                           */
/*                                                        */
/*  Version 2.1                                           */
/*  Martial MICHEL (AutoMap & AutoLink :                  */
/*                    Concept & Programming)              */
/*  & Delphine GOUJON (AutoLink : Concept)                */
/*    July 1998                                           */
/*                                                        */
/*  Version 2.11                                          */
/*  Martial MICHEL                                        */
/*  & Thibaut ROUFFINEAU (Testing)                        */
/*    November 1998                                       */
/*                                                        */
/*  Version 2.11.1                                        */
/*  Martial MICHEL                                        */
/*    January 1999                                        */
/*                                                        */
/*  Version 2.11.2                                        */
/*  Martial MICHEL                                        */
/*    February 1999                                       */
/*                                                        */
/*  Version 2.20                                          */
/*  Martial MICHEL                                        */
/*    June 1999                                           */
/*                                                        */
/*  Version 3.00b1                                        */
/*  Martial MICHEL                                        */
/*  & Olivier NICOLAS (Testing & AutoMap : Concept)       */
/*    April 2000                                          */
/*                                                        */
/*  Version 3.00b2                                        */
/*  Martial MICHEL                                        */
/*  & Olivier NICOLAS (Stress Testing)                    */
/*    August 2000                                         */
/*                                                        */
/**********************************************************/

/************************************************************/
/*
  Contains : AutoLink (v3.00 [beta 2]) functions and sub-functions
  Organization : NIST
  Developped for : use with AutoMap / AutoLink v3.00
  by  : Martial MICHEL (Guest Researcher at NIST / PhD student)
  Project Leader : Judith Ellen Devaney
  
  Comments or ideas are welcomed, please E-mail to :
  martial.michel@nist.gov
*/
/************************************************************/

#if 0
static char _autolink_inc_[] = 
"@(#)[v3.00 beta 2] AutoLink functions and sub-functions";
#endif

/*
  This is to avoid multiple definitions of some global variables 
 (including arrays of pointer to functions)
*/
#define __AL_AUTOLINK_INC__

/**********************************************************/
/*     Include AutoLink Internals Prototypes              */
/**********************************************************/

#include "autolink.h"

/* Version number check : number * 100 (need integer) */
#if ((! defined AMVERSION) || (! defined AMVERSION_BETA) || (AMVERSION != 300) || (AMVERSION_BETA != 2))
#error "ERROR : This version of AutoLink is designed to work only with entries generated from AutoMap v3.00 (beta 2)"
#endif

/**********************************************************/
/* AutoLink Debugging Internal Use                        */
/**********************************************************/
#include <stdlib.h>
#include <stdio.h>

#ifdef _AL_DEBUG_
int __AL_rank;
int __AL_log;
FILE *__AL_file;
#if (_AL_DBG_LVL & _AL_DBG_FLUSH)
#define _AL_ff fflush(__AL_file)
#else
#define _AL_ff
#endif

#if (_AL_DBG_LVL & (_AL_DBG_BENCH | _AL_DBG_UBENCH))
double __AL_starttime, __AL_endtime;
int __AL_PacketSize;
#endif

#endif

/**********************************************************/
/*     Include mpitypes.inc                               */	
/**********************************************************/

#include  "mpitypes.inc"

/**********************************************************/
/*     Global Variables                                   */	
/**********************************************************/

AL_Blocks*         AL_DataBlocks[_AL_TpUsed];
/* Memory blocks */

int                AL_PacketSize[_AL_TpUsed];
/* Max number of elements per packet */

AL_PendingIRecvRL* AL_PendingIRecvRLHandler;
/* Used to handle pending IRecv */

AL_PIRRLList    *AL_PIRRLList_freeList;
/* Used to keep a list of freeList for pending IRecv */

int  _AL_PacketChange;
/* to check status of user capability to change PacketSize */

int  _AL_ReturnCode;
/* is used for passing MPI error code to AL_Send and AL_Recv output */

int  _AL_MaxRanks;
/* to have the maximum number of ranks available */

MPI_Datatype  MPI_AL_InitMsg;
/* The InitMsg data-type for MPI */

/**********************************************************/
/*     Include al_routines.inc                            */	
/**********************************************************/

#include  "al_routines.inc"

/**********************************************************/
/*     User function to add a line to the AL log          */
/*     (will only produce entries if _AL_DEBUG is on)     */
/**********************************************************/

int
AL_LogEntry(char *entry)
{ /* AL_LogEntry */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_UBENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALUBM %f : %d] %s\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize, entry);_AL_ff;
#endif
#endif

  return(AL_True);
}

/**********/

void
_AL_LogEntry(char *entry)
{ /* _AL_LogEntry */
#ifdef _AL_DEBUG_
  fprintf(__AL_file, "[%.4d/%.8d] %s\n", __AL_rank, __AL_log++, entry);_AL_ff;
#endif
}

/********************************************************/
/*     AutoLink Internal MPI defined datatype           */
/********************************************************/

void
_AL_CreateInitMsgDT ()
{ /* _AL_CreateInitMsgDT */
  /* Derivated from an AutoMap created entry */
  AL_InitMsg   init;
  MPI_Aint     disp[2];
  MPI_Datatype type[2];
  int          blocklen[2];
  int          base;
  int          i;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_CreateInitMsgDT\n", __AL_rank, __AL_log++);_AL_ff;
#endif
#endif

  /* Set the displacement */
  MPI_Address(init.DataCount, &disp[0]);
  MPI_Address(&init.InitialDatatype, &disp[1]);
  /* Set a datatype */
  type[0] = MPI_UNSIGNED_LONG;
  type[1] = MPI_INT;
  /* Set datatype sizes */
  blocklen[0] = _AL_TpUsed;
  blocklen[1] = 1;
  /* Arrange displacement */
  base = disp[0];
  for (i = 0; i < 2; i++)
    disp[i] -= base;

  /* Create Data-type */
  MPI_Type_struct(2, blocklen, disp, type, &MPI_AL_InitMsg);
  MPI_Type_commit(&MPI_AL_InitMsg);
}

void
_AL_DeleteInitMsgDT ()
{ /* _AL_DeleteInitMsgDT */
  MPI_Type_free(&MPI_AL_InitMsg);
}

/**********************************************************/
/*     Error Handler                                      */
/**********************************************************/

#define _AL_ERROR_ MPI_ERR_OTHER

int
_AL_ErrHandler(int ret)
{ /* _AL_ErrHandler */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_ErrHandler : [Return Code : %d]\n", __AL_rank, __AL_log++, ret);_AL_ff;
#endif
#endif
  
  if (ret != MPI_SUCCESS) {
    _AL_ReturnCode = ret;
    
    return (AL_False);
  }
  
  return (AL_True);
}

/********************************************************/
/*     Datatype Traversal (Send - Part 1)               */
/********************************************************/

void
_AL_DataTraverse(void* buf, int datatype, AL_Internals *ALint)
{ /* _AL_DataTraverse */
  int tmp;
  /* To store datatype from NEXT ("cn" for "current node") */
  /* To store position in ADDRESS from NEXT */
  /* To store pointer to object in ADDRESS from NEXT */
  int   cn_dt, cn_pos;
  void* cn_ptr;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_DataTraverse \n", __AL_rank, __AL_log++); _AL_ff;
#endif
#if (_AL_DBG_LVL & _AL_DBG_DATA)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_DataTraverse : Add Address of Entry Node in Address\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  /* Add Address of Entry Node in Address */
  tmp = AL_SArray_AddLast(&(ALint->AL_Address), datatype, buf);
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_DataTraverse : Add Entry Node in NEXT\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  /* Add Entry Node in NEXT */
  AL_Queue_Add(ALint->AL_Next, datatype, tmp);
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_DataTraverse : Mark Entry Node\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  /* Mark Entry Node */
  AL_HT_Add(ALint->AL_Mark, (unsigned long int) buf, tmp);
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_DataTraverse : While there are elements to visit in NEXT\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] _AL_DataTraverse : Main Loop Start\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif

  /* While there are elements to visit in NEXT */
  while (AL_Queue_IsEmpty(ALint->AL_Next) == AL_False) {
    
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
    fprintf(__AL_file, "[%.4d/%.8d] _AL_DataTraverse : Reach current node in NEXT\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

    /* Reach current node in NEXT */
    cn_dt  = AL_Queue_GetDatatype(ALint->AL_Next);
    cn_pos = AL_Queue_GetPosition(ALint->AL_Next);
    cn_ptr = AL_SArray_GetValue(ALint->AL_Address, cn_dt, cn_pos);

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
    fprintf(__AL_file, "[%.4d/%.8d] _AL_DataTraverse : Current Node datatype = %d / Position : %d / Pointer : 0X%.8x\n", __AL_rank, __AL_log++, cn_dt, cn_pos, (int) cn_ptr); _AL_ff;
#endif
#endif
    
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
    fprintf(__AL_file, "[%.4d/%.8d] _AL_DataTraverse : For each children of current node : [Current Node Datatype : %d] [CN Address :  0X%.8X]\n", __AL_rank, __AL_log++, cn_dt, (int) cn_ptr); _AL_ff;
#endif
#endif

    /* For each children of current node "Traverse" */
    _AL_Traverse[cn_dt](cn_ptr, ALint);
    
    /* Inc number of elements for this data-types */
    (ALint->AL_DataCount[cn_dt])++;
    
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
    fprintf(__AL_file, "[%.4d/%.8d] _AL_DataTraverse : Goto next in NEXT\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
    
    /* Goto next in NEXT */
    AL_Queue_Remove(ALint->AL_Next);
  }
}

/********************************************************/
/*     Memory Block control                             */
/********************************************************/

void
_AL_MemBlocksFree()
{ /* _AL_MemBlocksFree */
  int i;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_MemBlocksFree\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  for (i = 0; i < _AL_TpUsed; i++)
    /*
      Remaining is left to the user; if during AL_Recv 
      rqlist != NULL, user has to delete the data-type
      rqlist == NULL, user has nothing to worry about
    */
    AL_Blocks_Delete(&AL_DataBlocks[i]);
}

/********************************************************/
/*     Packet Size operations                           */
/********************************************************/

void
_AL_CalculatePacketSize(int size)
{ /* _AL_CalculatePacketSize */
  int i;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_CalculatePacketSize : [Size: %d]\n", __AL_rank, __AL_log++, size); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  __AL_PacketSize = size;
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] _AL_CalculatePacketSize\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif
  
  /* Note : No verification on "user control" */

  for ( i = 0; i < _AL_TpUsed; i++) {
    _AL_SetPacketSize[i](size);
    if (AL_PacketSize[i] == 0)
      AL_PacketSize[i] = 1;
  }
}

/****************************************/

int
AL_SetPacketSize(int size)
{ /* AL_SetPacketSize */
  int i;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_SetPacketSize : [Size: %d]\n", __AL_rank, __AL_log++, size); _AL_ff;
#endif
#endif

  /* Users are allowed to change packet size */
  if (_AL_PacketChange == AL_True) {

#ifdef _AL_DEBUG_
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
    __AL_endtime = MPI_Wtime();
    __AL_PacketSize = size;
    fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_SetPacketSize\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif
  
    /* Let's free old memory blocks */
    _AL_MemBlocksFree();

    /* Readapt the array */
    _AL_CalculatePacketSize(size);

    /* Recreate new memory blocks */
    for(i = 0; i < _AL_TpUsed; i++) {

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
      fprintf(__AL_file, "[%.4d/%.8d] AL_SetPacketSize : Creating Memory blocks (%d)\n", __AL_rank, __AL_log++, i); _AL_ff;
#endif
#endif

      _AL_PrepareBuffer[i]();
    }
    
    return((int) AL_True);
  } else {
    /* Users are not allowed to change packet size */

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
    fprintf(__AL_file, "[%.4d/%.8d] AL_SetPacketSize : REFUSED\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

    return ((int) AL_False);
  }
}

/********************************************************/
/*     Init/Finalize operations                         */
/********************************************************/

void
AL_Free(AL_ReqList** rqlist)
{ /* AL_Free */
  AL_ReqList_Delete(*rqlist);

  *rqlist = NULL;
}

/************************************************************/

void
AL_Init(int packetsize)
{ /* AL_Init */
  int	i, psize;
  
#ifdef _AL_DEBUG_
#ifdef _AL_DEBUG_FILENAME
  char __AL_fname[256];
#endif
  __AL_log = 0;
  MPI_Comm_rank(MPI_COMM_WORLD, &__AL_rank);
#ifdef _AL_DEBUG_FILENAME
  sprintf(__AL_fname, "%s.%d", _AL_DEBUG_FILENAME, __AL_rank);
  if ((__AL_file = fopen((const char *) __AL_fname, "w")) == NULL) {
    fprintf(stderr,
	    "\nAutoLink -- Internal error : Could not create log file (%s)\n",
	    __AL_fname);
    exit(EXIT_FAILURE);
  }
#else
  __AL_file = stderr;
#endif
#if (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_starttime = MPI_Wtime();
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_Init\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif 
#endif
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Init : Initialize the functions array we are about to use\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  /* First let's get the max ranks number */
  MPI_Comm_size(MPI_COMM_WORLD, &_AL_MaxRanks);

  /* Initialize the functions array we are about to use */
  _AL_InitFunctions();
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Init : Create the MPI types before using some\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  /* Create the MPI types before using some */
  Build_MPI_Types();
  _AL_CreateInitMsgDT();

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Init : Set Packet Size\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  /* Set Packet Size */
  psize = (packetsize >= 0) ? packetsize : AL_DEFAULTMAXPACKET;
  _AL_CalculatePacketSize(psize);

#ifndef AL_USERCONTROL_PACKET
  _AL_PacketChange = AL_False;
#else
  _AL_PacketChange = AL_True;
#endif

  /* Create Memory blocks */
  for(i = 0; i < _AL_TpUsed; i++) {

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_Init : Creating Memory blocks (%d)\n", __AL_rank, __AL_log++, i); _AL_ff;
#endif
#endif

    _AL_PrepareBuffer[i]();
  }

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Init : Creating Pending IRecv handlers\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  
  /* Create Pending IRecv handler */
  AL_PendingIRecvRLHandler = AL_PendingIRecvRL_Create();
  AL_PIRRLList_freeList = NULL;
  AL_PIRRLList_AddFreeElements();
}

/************************************************************/

void
AL_Finalize()
{ /* AL_Finalize */

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Finalize\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_Finalize\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Finalize : Clearing allocated memory\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  _AL_MemBlocksFree();

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Finalize : Remove MPI created data-types\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  Free_MPI_Types();
  _AL_DeleteInitMsgDT();

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Finalize : Deleting Pending IRecv handler\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  
  /* Delete Pending IRecv handler */
  AL_PendingIRecvRL_Delete(AL_PendingIRecvRLHandler);
  AL_PIRRLList_DeleteFreeElemenst();

#ifdef _AL_DEBUG_
  fprintf(__AL_file, "[%.4d/%.8d] AL_Finalize : AutoLink finished ... closing log\n", __AL_rank, __AL_log++); _AL_ff;
#ifdef _AL_DEBUG_FILENAME
  fclose(__AL_file);
#endif
#endif
}

/********************************************************/
/*     Blocking send operations                         */
/********************************************************/

int
_AL_DataSend(int dest, int tag, MPI_Comm comm,
	     int datatype, AL_Internals *ALint)
{ /* _AL_DataSend */

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_DataSend : [Dest Rank : %d] [Tag : %d]\n", __AL_rank, __AL_log++, dest, tag); _AL_ff;
#endif
#if (_AL_DBG_LVL & _AL_DBG_FINT & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] _AL_DataSend Start\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif
  
  if ((int) ALint->AL_DataCount[datatype] > 0) {

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_STEP)
    fprintf(__AL_file, "[%.4d/%.8d] _AL_DataSend : Will SEND AL_Data[%d]\n", __AL_rank, __AL_log++, datatype);
#endif
#endif
    
    if (_AL_ErrHandler(
		       MPI_Send(ALint->AL_Data[datatype],
				ALint->AL_DataCount[datatype], 
				_AM_Types[datatype], dest, tag, comm) 
		       ) == AL_False
	) {
      
      return (AL_False);
    }
  }
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_DataSend : return (AL_True)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#if (_AL_DBG_LVL & _AL_DBG_FINT & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] _AL_DataSend Finish\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif

  return (AL_True);
}

/****************************************/

void
_AL_SendMalloc(AL_Internals *ALint)
{ /* _AL_SendMalloc */
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_SendMalloc\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  ALint->AL_Next = AL_Queue_Create();
  ALint->AL_Mark = AL_HT_Create();
  ALint->AL_Address = AL_SArray_Create();
}

/****************************************/

void
_AL_SendFree(AL_Internals *ALint)
{ /* _AL_SendFree */

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_SendFree\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  AL_Queue_Delete(ALint->AL_Next);
  AL_HT_Delete(&(ALint->AL_Mark));
  AL_SArray_Delete(ALint->AL_Address);
}

/****************************************/

void
_AL_SendErrorFree(AL_Internals *ALint)
{ /* _AL_SendErrorFree */

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_SendErrorFree\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  
  _AL_SendFree(ALint);
}

/****************************************/

int
AL_Send(void* buf, int datatype, int dest, int tag, MPI_Comm comm)
{ /* AL_Send */

  int i, j;
  /* AutoLink internals */
  AL_Internals ALint;
  AL_InitMsg   InitMsg;
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Send : [Address of Entry Node : 0X%.8X] [Datatype Number of EN : %d] [Destination Rank : %d] [Tag : %d]\n", __AL_rank, __AL_log++, (int) buf, datatype, dest, tag); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_Send Start\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Send : Create AL internals needed\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  /* Create AL internals needed */
  _AL_SendMalloc(&ALint);

  for(i = 0; i < _AL_TpUsed; i++) {
    InitMsg.DataCount[i] = 0;
    ALint.AL_DataCount[i] = 0;
    _AL_GetBlockBuffer[i](&ALint);
  }

  _AL_DataTraverse(buf, datatype, &ALint);
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Send : Acutal Send\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_Send : Main Loop Finish\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif

  /* Create InitMsg */
  for (i = 0; i < _AL_TpUsed; i++) {
    InitMsg.DataCount[i] = ALint.AL_DataCount[i];
    /* We need empty number of elements to be able to add some later */
    ALint.AL_DataCount[i] = 0;
  }
  InitMsg.InitialDatatype = datatype;
  
  /* Send InitMsg */
  if (_AL_ErrHandler(MPI_Send(&InitMsg, 1, MPI_AL_InitMsg, dest, tag, comm))
      == AL_False) {

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_Send : Error while sending InitMsg\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

    _AL_SendErrorFree(&ALint);
    for (i = 0; i < _AL_TpUsed; i++)
      _AL_GiveBlockBuffer[i](&ALint);

    return (_AL_ReturnCode);
  }
  
  /* Copy buffers and send when one is full */
  for (i = 0; i < _AL_TpUsed; i++) {
	  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_Send : Sending datatype %d\n", __AL_rank, __AL_log++, i); _AL_ff;
#endif
#endif

    for (j = 0; j < (int) InitMsg.DataCount[i]; j++) {
      if (_AL_CopyAdapt[i](AL_SArray_GetValue(ALint.AL_Address, i, j), &ALint)
	  == AL_PacketSize[i]) {
	if (_AL_DataSend(dest, tag, comm, i, &ALint) == AL_False) {
	  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
	  fprintf(__AL_file, "[%.4d/%.8d] AL_Send : Error while sending\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
	  
	  _AL_SendErrorFree(&ALint);
	  for (i = 0; i < _AL_TpUsed; i++)
	    _AL_GiveBlockBuffer[i](&ALint);
	  
	  return (_AL_ReturnCode);
	}

	ALint.AL_DataCount[i] = 0;
      }
    }
    
    /* Let's send left over */
    if (_AL_DataSend(dest, tag, comm, i, &ALint) == AL_False) {
      
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
      fprintf(__AL_file, "[%.4d/%.8d] AL_Send : Error while sending\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
      
      _AL_SendErrorFree(&ALint);
      for (i = 0; i < _AL_TpUsed; i++)
	_AL_GiveBlockBuffer[i](&ALint);
      
      return (_AL_ReturnCode);
    }
  }

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Send : Delete AL internals needed\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  /* Delete AL internals needed */
  _AL_SendFree(&ALint);
  for (i = 0; i < _AL_TpUsed; i++)
    _AL_GiveBlockBuffer[i](&ALint);
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Send : return (MPI_SUCCESS)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_Send Finish\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif

  return (MPI_SUCCESS);
}

/********************************************************/
/*     Blocking Receive operation                       */
/********************************************************/

void
_AL_RecvMalloc(AL_Internals *ALint)
{ /* _AL_RecvMalloc */
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_RecvMalloc\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  ALint->AL_Address = AL_SArray_Create();
}

/****************************************/

void
_AL_RecvFree(AL_Internals *ALint)
{ /* _AL_RecvFree */
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_RecvFree\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  AL_SArray_Delete(ALint->AL_Address);
}

/****************************************/

void
_AL_RecvErrorFree(AL_Internals *ALint)
{ /* _AL_RecvErrorFree */
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_RecvErrorFree\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  
  _AL_RecvFree(ALint);
}

/****************************************/

int
AL_Recv(void** buf, int* datatype, int source, int tag, 
	MPI_Comm comm, MPI_Status* status, AL_ReqList** rqlist)
{ /* AL_Recv */
  int i, j;

  /* AutoLink internals */
  AL_Internals ALint;
  AL_InitMsg   InitMsg;
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Recv : [Source Rank : %d] [Tag : %d]\n", __AL_rank, __AL_log++, source, tag); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_Recv Start\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Recv : Create AL internals needed\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  /* Create AL internals needed */
  _AL_RecvMalloc(&ALint);

  for(i = 0; i < _AL_TpUsed; i++) {
    ALint.AL_DataCount[i] = 0;
    if (rqlist == NULL)
      _AL_GetBlockBuffer[i](&ALint);
  }
  
  /* Create a request list if required */
  if ((rqlist != NULL) && (*rqlist == NULL))
    *rqlist = (AL_ReqList *) AL_ReqList_Create();
  

  /* Receive InitMsg */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Recv : Recv Number of data to receive\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  if (_AL_ErrHandler(MPI_Recv(&InitMsg, 1, MPI_AL_InitMsg,
			      source, tag, comm, status)
		     ) == AL_False) {

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_Recv : Error while receiving\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

    _AL_RecvErrorFree(&ALint);
    for (i = 0; i < _AL_TpUsed; i++)
      _AL_GiveBlockBuffer[i](&ALint);
      
    return (_AL_ReturnCode);
  }

  /* Receive Packets and create them in memory */
  for (i = 0; i < _AL_TpUsed; i++) {
    ALint.AL_DataCount[i] = InitMsg.DataCount[i];

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_Recv : Receiving datatype %d (%d elements)\n", __AL_rank, __AL_log++, i, (int) InitMsg.DataCount[i]); _AL_ff;
#endif
#endif
    
    while ((int) InitMsg.DataCount[i] > 0) {
      int tmpdt;

      /* Get data from source */
      if ((int) InitMsg.DataCount[i] >= AL_PacketSize[i])
	tmpdt = AL_PacketSize[i];
      else
	tmpdt = InitMsg.DataCount[i];
      
      /* Create a new buffer if required */
      if (rqlist != NULL)
	if (tmpdt > 0)
	  _AL_GetBlockBuffer[i](&ALint);
      
      /* Receive Packet */
      if (_AL_ErrHandler(MPI_Recv(ALint.AL_Data[i], tmpdt, 
				  _AM_Types[i], source, tag, comm, status)
			 ) == AL_False
	  ) {
	
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
	fprintf(__AL_file, "[%.4d/%.8d] AL_Recv : Error while receiving\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
	
	_AL_RecvErrorFree(&ALint);
	for (i = 0; i < _AL_TpUsed; i++)
	  _AL_GiveBlockBuffer[i](&ALint);
	
	return (_AL_ReturnCode);
      }
      
      /* Add data to Request List if required */
      if (rqlist != NULL) 
	AL_ReqList_Add(rqlist, ALint.AL_Data[i], i);
      
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
      fprintf(__AL_file, "[%.4d/%.8d] AL_Recv : For each element in PACKET\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

      /* For each element in PACKET */
      for (j = 0; j < tmpdt; j++) {
	
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
	fprintf(__AL_file, "[%.4d/%.8d] AL_Recv : Create element in memory (%d, %d)\n", __AL_rank, __AL_log++, i, j); _AL_ff;
#endif
#endif

	/* Create element in memory 
	   Add address of created element in ADDRESS */

	/*
	  Two methods of creation :
	  - Using one block and splitting it (rqlist != NULL)
	  - Creating each element on its own (rqlist == NULL)
	*/
	if (rqlist == NULL)
	  AL_SArray_AddLast(&(ALint.AL_Address), i, 
			    _AL_MemCreate[i](i, j, &ALint));
	else 
	  AL_SArray_AddLast(&(ALint.AL_Address), i, 
			    _AL_GetBlockElement[i](j, &ALint));
      }
	
      /* Less to receive */
      InitMsg.DataCount[i] -= tmpdt;
    }
  }
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Recv : For each elements of LINKS ...\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_Recv : For each Start\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif

  /* For each elements of LINKS ... */
  for (i = 0; i < _AL_TpUsed; i++) {
    for (j = 0; j < (int) ALint.AL_DataCount[i]; j++) {

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
      fprintf(__AL_file, "[%.4d/%.8d] AL_Recv : Will run _AL_Restore[%d] for AL_Address(%d, %d)\n", __AL_rank, __AL_log++, i, i, j); _AL_ff;
#endif
#endif

      _AL_Restore[i](AL_SArray_GetValue(ALint.AL_Address, i, j), &ALint);
    }
  }

#ifdef _AL_DEBUG_
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_Recv : For each Finish\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif  
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Recv : Result is Intial Object with recreated links\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_Recv : Main Loop Finish\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif

  /* Result is Intial Object with recreated links */
  *buf = AL_SArray_GetValue(ALint.AL_Address, InitMsg.InitialDatatype, 0);
  *datatype = InitMsg.InitialDatatype;
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Recv : Delete AL internals needed\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  /* Delete AL internals needed */
  _AL_RecvFree(&ALint);
  if (rqlist == NULL)
    for (i = 0; i < _AL_TpUsed; i++)
      _AL_GiveBlockBuffer[i](&ALint);
  
#ifdef _AL_DEBUG_
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_Recv Finish\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_Recv : return (MPI_SUCCESS)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  return (MPI_SUCCESS);
}

/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/
/************************************************************/


/************************************************************/
/*     Non Blocking operations                              */
/************************************************************/

MPI_Request *
_AL_newMPIRequest()
{ /* _AL_newMPIRequest */
  MPI_Request *tmp;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_newMPIRequest\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  tmp = malloc(sizeof(*tmp));

  return (tmp);
}

/****************************************/

void
_AL_ISendMalloc(AL_Internals *ALint)
{ /* _AL_ISendMalloc */
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_ISendMalloc\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  ALint->AL_Next = AL_Queue_Create();
  ALint->AL_Mark = AL_HT_Create();
  ALint->AL_Address = AL_SArray_Create();
}

/****************************************/

int
_AL_IDataSend(int dest, int tag, MPI_Comm comm,
	      int datatype, AL_ISendRL* rlist)
{ /* _AL_IDataSend */
  MPI_Request *tmprq;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_IDataSend : [Dest Rank : %d] [Tag : %d]\n", __AL_rank, __AL_log++, dest, tag); _AL_ff;
#endif
#if (_AL_DBG_LVL & _AL_DBG_FINT & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] _AL_IDataSend Start\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif
  
  if ((int) rlist->ALint->AL_DataCount[datatype] > 0) {

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_STEP)
    fprintf(__AL_file, "[%.4d/%.8d] _AL_IDataSend : Will SEND AL_Data[%d] (%d elements)\n", __AL_rank, __AL_log++, datatype, (int) rlist->ALint->AL_DataCount[datatype]);
#endif
#endif
    tmprq = _AL_newMPIRequest();

    if (_AL_ErrHandler(
		       MPI_Isend(rlist->ALint->AL_Data[datatype],
				 rlist->ALint->AL_DataCount[datatype], 
				 _AM_Types[datatype], dest, tag, comm,
				 tmprq) 
		       ) == AL_False
	) {
      
      return (AL_False);
    }

    AL_IIRL_AddLast(rlist->IIRL, datatype, tmprq,
		    rlist->ALint->AL_Data[datatype]);
  }
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_IDataSend : return (AL_True)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#if (_AL_DBG_LVL & _AL_DBG_FINT & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] _AL_IDataSend Finish\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif

  return (AL_True);
}

/********************/

int
AL_ISendTest(AL_ISendRL **rlist)
{ /* AL_ISendTest */

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_ISendTest [rlist : %X] [State : %d]\n", __AL_rank, __AL_log++, *rlist, (*rlist)->State); _AL_ff;
#endif
#endif

  /* Pre-cond : we have a rlist to rely on */
  if (*rlist == NULL)
    return (_AL_ERROR_);

  /* First let's test the InitMsg */
  if ((*rlist)->State == AL_ISendRL_State_WAITIM) {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_ISendTest : Test InitMsg\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
    if (AL_ISendRL_TestInitMsg(*rlist) == AL_True)
      (*rlist)->State = AL_ISendRL_State_WAITMSG;
    else
      return (_AL_ERROR_);
  }


  /* Then let's test all the remaining messages */
  if ((*rlist)->State == AL_ISendRL_State_WAITMSG) {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_ISendTest : Test all messages\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
    if (AL_IIRL_Testall((*rlist)->IIRL) != AL_True)
      return (_AL_ERROR_);

    (*rlist)->State = AL_ISendRL_State_DONE;
  }

  if ((*rlist)->State != AL_ISendRL_State_DONE)
    return (_AL_ERROR_);

  /* Everything went smoothly */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_ISendTest : return (MPI_SUCCESS)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  return (MPI_SUCCESS);
}

/********************/

int
AL_ISendWait(AL_ISendRL **rlist)
{ /* AL_ISendWait */

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_ISendWait [rlist : %X] [State : %d]\n", __AL_rank, __AL_log++, *rlist, (*rlist)->State); _AL_ff;
#endif
#endif

  /* Pre-cond : we have a rlist to rely on */
  if (*rlist == NULL)
    return (_AL_ERROR_);

  /* First let's wait on the InitMsg */
  if ((*rlist)->State == AL_ISendRL_State_WAITIM) {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_ISendWait : Wait on InitMsg\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
    if (AL_ISendRL_WaitInitMsg(*rlist) == AL_True)
      (*rlist)->State = AL_ISendRL_State_WAITMSG;
    else
      return (_AL_ERROR_);
  }

  /* Then let's wait on all the remaining messages */
  if ((*rlist)->State == AL_ISendRL_State_WAITMSG) {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_ISendWait : Wait on all messages\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
    if (AL_IIRL_Waitall((*rlist)->IIRL) != AL_True)
      return (_AL_ERROR_);

    (*rlist)->State = AL_ISendRL_State_DONE;
  }

  if ((*rlist)->State != AL_ISendRL_State_DONE)
    return (_AL_ERROR_);

  /* Everything went smoothly */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_ISendWait : return (MPI_SUCCESS)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  return (MPI_SUCCESS);
}

/********************/

int
AL_ISendComplete(AL_ISendRL **rlist)
{ /* AL_ISendComplete */
  int tmp;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_ISendComplete [rlist : %X] [State : %d]\n", __AL_rank, __AL_log++, *rlist, (*rlist)->State); _AL_ff;
#endif
#endif

  /* Pre-cond : we have a rlist to rely on */
  if (*rlist == NULL)
    return (_AL_ERROR_);

  /* First, let's make sure we have finished */
  if ((*rlist)->State != AL_ISendRL_State_DONE) {
    tmp = AL_ISendWait(rlist);

    if (tmp != MPI_SUCCESS)
      return (tmp);
  }

  /* Then let's give back all the used blocks */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_ISendComplete : Give back all used blocks\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  AL_ISendRL_GiveBlockBuffer(*rlist);

  /* Finally, let's empty the Pointer */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_ISendComplete : Cleaning memory\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  _AL_SendErrorFree((*rlist)->ALint);
  AL_ISendRL_Delete(*rlist);
  *rlist = NULL;
  
  /* Everything went smoothly */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_ISendComplete : return (MPI_SUCCESS)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  return (MPI_SUCCESS);
}

/****************************************/

int
AL_ISend(void* buf, int datatype, int dest, int tag, MPI_Comm comm,
	 AL_ISendRL** rlist)
{ /* AL_ISend */
  
  int i, j;
  MPI_Request *tmprq;

  /* AutoLink internals */
  AL_Internals  *ALint;
  AL_InitMsg    *InitMsg;
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_ISend : [Address of Entry Node : 0X%.8X] [Datatype Number of EN : %d] [Destination Rank : %d] [Tag : %d]\n", __AL_rank, __AL_log++, (int) buf, datatype, dest, tag); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_ISend Start\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_ISend : Create AL internals needed\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  /* Create AL internals needed */
  ALint = (AL_Internals *) malloc(sizeof(AL_Internals));
  InitMsg = (AL_InitMsg *) malloc(sizeof(AL_InitMsg));

  _AL_ISendMalloc(ALint);
  
  for(i = 0; i < _AL_TpUsed; i++)
    ALint->AL_DataCount[i] = 0;

  _AL_DataTraverse(buf, datatype, ALint);
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_ISend : Acutal Send\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_ISend : Main Loop Finish\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif

  /* Create the user reqlist */
  *rlist = AL_ISendRL_Create(buf, datatype, dest, tag, comm, ALint, InitMsg);

  /* Create InitMsg */
  for (i = 0; i < _AL_TpUsed; i++) {
    InitMsg->DataCount[i] = ALint->AL_DataCount[i];
    /* We need empty number of elements to be able to add some later */
    ALint->AL_DataCount[i] = 0;
  }
  InitMsg->InitialDatatype = datatype;

  tmprq = _AL_newMPIRequest();

  /* Send InitMsg */
  if (_AL_ErrHandler(MPI_Isend(InitMsg, 1, MPI_AL_InitMsg, dest, tag, comm,
			       tmprq))
      == AL_False) {

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_ISend : Error while sending InitMsg\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

    _AL_SendErrorFree(ALint);
    for (i = 0; i < _AL_TpUsed; i++)
      _AL_GiveBlockBuffer[i](ALint);

    return (_AL_ERROR_);
  }

  /* Set InitMsg */
  AL_ISendRL_SetInitMsgRq(*rlist, tmprq);
  (*rlist)->State = AL_ISendRL_State_WAITIM;

  /* Copy buffers and send when one is full */
  for (i = 0; i < _AL_TpUsed; i++) {
    int getnewbuffer = AL_True;
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_ISend : Sending datatype %d\n", __AL_rank, __AL_log++, i); _AL_ff;
#endif
#endif

    for (j = 0; j < (int) InitMsg->DataCount[i]; j++) {
      /* We only create new buffer if we need to */
      if (getnewbuffer == AL_True) {
	_AL_GetBlockBuffer[i](ALint);
	getnewbuffer = AL_False;
      }

      if (_AL_CopyAdapt[i](AL_SArray_GetValue(ALint->AL_Address, i, j), ALint)
	  == AL_PacketSize[i]) {
	if (_AL_IDataSend(dest, tag, comm, i, *rlist) == AL_False) {
	  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
	  fprintf(__AL_file, "[%.4d/%.8d] AL_ISend : Error while sending\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
	  
	  _AL_SendErrorFree(ALint);
	  for (i = 0; i < _AL_TpUsed; i++)
	    _AL_GiveBlockBuffer[i](ALint);
	  
	  return (_AL_ERROR_);
	}

	ALint->AL_DataCount[i] = 0;
	getnewbuffer = AL_True;
      }
    }
    
    /* Let's send left over */
    if (_AL_IDataSend(dest, tag, comm, i, *rlist) == AL_False) {
      
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
      fprintf(__AL_file, "[%.4d/%.8d] AL_ISend : Error while sending\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
      
      _AL_SendErrorFree(ALint);
      for (i = 0; i < _AL_TpUsed; i++)
	_AL_GiveBlockBuffer[i](ALint);
      
      return (_AL_ERROR_);
    }
  }

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_ISend : [rlist : %X] [State : %d]\n", __AL_rank, __AL_log++, *rlist, (*rlist)->State); _AL_ff;
  fprintf(__AL_file, "[%.4d/%.8d] AL_ISend : return (MPI_SUCCESS)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_ISend Finish\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif

  return (MPI_SUCCESS);
}

/************************************************************/

void
_AL_IRecvMalloc(AL_Internals *ALint)
{ /* _AL_IRecvMalloc */
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_IRecvMalloc\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  ALint->AL_Address = AL_SArray_Create();
}

/********************/

int
_AL_IDataRecv(AL_IRecvRL** rlist)
{ /* _AL_IRecvAddAll */
  int i;
  int DataCount[_AL_TpUsed];
  MPI_Request *tmprq;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_IDataRecv [rlist : %X]\n", __AL_rank, __AL_log++, *rlist); _AL_ff;
#endif
#endif

  /* Add Packets in Receive Request List  */
  for (i = 0; i < _AL_TpUsed; i++) {
    DataCount[i] = (*rlist)->InitMsg->DataCount[i];

    while (DataCount[i] > 0) {
      int tmpdt;

      /* Get data from source */
      if (DataCount[i] >= AL_PacketSize[i])
	tmpdt = AL_PacketSize[i];
      else
	tmpdt = DataCount[i];
      
      /* Get new buffers */
      _AL_GetBlockBuffer[i]((*rlist)->ALint);
      tmprq = _AL_newMPIRequest();
      
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
      fprintf(__AL_file, "[%.4d/%.8d] _AL_IRecvAddAll : Adding Request datatype %d (%d elements out of %d total)\n", __AL_rank, __AL_log++, i, tmpdt, (int) (*rlist)->InitMsg->DataCount[i]); _AL_ff;
#endif
#endif
      
      if (_AL_ErrHandler(MPI_Irecv((*rlist)->ALint->AL_Data[i], tmpdt, 
				   _AM_Types[i], (*rlist)->source,
				   (*rlist)->tag, (*rlist)->comm,
				   tmprq)
			 ) == AL_False
	  ) {
	
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
	fprintf(__AL_file, "[%.4d/%.8d] _AL_IRecvAddAll : Error while receiving\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
	
	for (i = 0; i < _AL_TpUsed; i++)
	  _AL_GiveBlockBuffer[i](&((*rlist)->ALint));
	
	return (AL_False);
      }
      
      /* Add data to ReqList */
      AL_IIRL_AddLast((*rlist)->IIRL, i, tmprq, (*rlist)->ALint->AL_Data[i]);

      /* Less to receive */
      DataCount[i] -= tmpdt;
    }
  }

  return(AL_True);
}

/********************/

int
_AL_IRecvProcessMessages(AL_IRecvRL **rlist)
{ /* _AL_IRecvProcessMessages */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_IRecvProcessMessages [rlist : %X]\n", __AL_rank, __AL_log++, *rlist); _AL_ff;
#endif
#endif

  if (AL_IRecvRL_TestInitMsg(*rlist) != AL_True)
    return (_AL_ERROR_);

  if (_AL_IDataRecv(rlist) != AL_True)
    return (_AL_ERROR_);
    
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_IRecvProcessMessages : return (MPI_SUCCESS)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  return (MPI_SUCCESS);
}

/********************/

int
_AL_IRecvProcessInitMsg(AL_IRecvRL **rlist)
{ /* _AL_IRecvProcessInitMsg */
  MPI_Request   *tmprq;
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_IRecvProcessInitMsg [rlist : %X]\n", __AL_rank, __AL_log++, *rlist); _AL_ff;
#endif
#endif

  /* Let's add the request for the InitMsg */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_IRecvProcessInitMsg : Recv Number of data to receive\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  
  tmprq = _AL_newMPIRequest();

  if (_AL_ErrHandler(MPI_Irecv((*rlist)->InitMsg, 1, MPI_AL_InitMsg,
			       (*rlist)->source, (*rlist)->tag, (*rlist)->comm,
			       tmprq)
		     ) == AL_False) {
    int i;
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvProcessInitMsg: Error while receiving\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
    
    for (i = 0; i < _AL_TpUsed; i++)
      _AL_GiveBlockBuffer[i](&((*rlist)->ALint));
    
    return (_AL_ERROR_);
  }
  
  /* Set InitMsgRq */
  AL_IRecvRL_SetInitMsgRq(*rlist, tmprq);

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvProcessInitMsg: return (MPI_SUCCESS)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  return (MPI_SUCCESS);
}

/********************/

void
_AL_IRecv_MemCreate(AL_IRecvRL **rlist)
{ /* _AL_IRecv_MemCreate */
  int i, j;
  int dt, tmpdt;
  int DataCount[_AL_TpUsed];

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_IRecv_MemCreate [rlist : %X]\n", __AL_rank, __AL_log++, *rlist); _AL_ff;
#endif
#endif

  for (i = 0; i < _AL_TpUsed; i++)
    DataCount[i] = (*rlist)->InitMsg->DataCount[i];
  
  for (i = 0; i < (*rlist)->IIRL->firstavailable; i++) {
    
    dt = (*rlist)->IIRL->datatype[i];
    (*rlist)->ALint->AL_Data[dt] = (*rlist)->IIRL->block[i];
    
    if (DataCount[dt] > AL_PacketSize[dt])
      tmpdt = AL_PacketSize[dt];
    else
      tmpdt = DataCount[dt];
    
    /* Set the rqlist */
    if ((*rlist)->rqlist != NULL)
      AL_ReqList_Add((*rlist)->rqlist, (*rlist)->ALint->AL_Data[dt], dt);

    /* For each element in PACKET */
    for (j = 0; j < tmpdt; j++) {

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
      fprintf(__AL_file, "[%.4d/%.8d] AL_IRecv_MemCreate : Create element in memory (%d, %d)\n", __AL_rank, __AL_log++, dt, j); _AL_ff;
#endif
#endif
      
      /* Create element in memory 
	 Add address of created element in ADDRESS */
      
      /*
	Two methods of creation :
	- Using one block and splitting it (rqlist != NULL)
	- Creating each element on its own (rqlist == NULL)
      */
      if ((*rlist)->rqlist == NULL)
	AL_SArray_AddLast(&((*rlist)->ALint->AL_Address), dt, 
			  _AL_MemCreate[dt](dt, j, ((*rlist)->ALint)));
      else 
	AL_SArray_AddLast(&((*rlist)->ALint->AL_Address), dt, 
			  _AL_GetBlockElement[dt](j, ((*rlist)->ALint)));
    }
    
    /* Less to receive */
    DataCount[dt] -= tmpdt;
  }

}

/********************/

void
_AL_IRecvLink(AL_IRecvRL **rlist)
{ /* _AL_IRecvLink */
  int i ,j;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_IRecvLink [rlist : %X]\n", __AL_rank, __AL_log++, *rlist); _AL_ff;
#endif
#endif
  
  /* For each elements of LINKS ... */
  for (i = 0; i < _AL_TpUsed; i++) {
    for (j = 0; j < (int) (*rlist)->InitMsg->DataCount[i]; j++) {

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
      fprintf(__AL_file, "[%.4d/%.8d] _AL_IRecvLink : Will run _AL_Restore[%d] for AL_Address(%d, %d)\n", __AL_rank, __AL_log++, i, i, j); _AL_ff;
#endif
#endif

      _AL_Restore[i](AL_SArray_GetValue((*rlist)->ALint->AL_Address, i, j),
		     ((*rlist)->ALint));
    }
  }
}

/********************/

int
_AL_IRecvTreatPending(AL_IRecvRL **rlist)
{ /* _AL_IRecvTreatPending */
  AL_IRecvRL *tmp;
  int doit = AL_True;
  int alldone = AL_False;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
  int inc = 0;
#endif
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_IRecvTreatPending [rlist : %X]\n", __AL_rank, __AL_log++, *rlist); _AL_ff;
#endif
#endif
  
  while (doit == AL_True) {
    doit = AL_False;
    
    tmp = AL_PendingIRecvRL_GetPending(&AL_PendingIRecvRLHandler,
				       (*rlist)->realrank, (*rlist)->tag);
    if (tmp != NULL) {
      
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
      inc++;
      fprintf(__AL_file, "[%.4d/%.8d] _AL_IRecvTreatPending : Will process %X on State %d\n", __AL_rank, __AL_log++, tmp, tmp->State); _AL_ff;
#endif
#endif
      
      if (tmp->State == AL_IRecvRL_State_PENDING) {
	if (_AL_IRecvProcessInitMsg(&tmp) == MPI_SUCCESS) {
	  tmp->State = AL_IRecvRL_State_WAITIM;
	  doit = AL_True;
	}
      }
      if (tmp->State == AL_IRecvRL_State_WAITIM) {
	doit = AL_False;
	if (_AL_IRecvProcessMessages(&tmp) == MPI_SUCCESS) {
	  tmp->State = AL_IRecvRL_State_WAITMSG;
	  doit = AL_True;
	}
      }
      if (tmp->State == AL_IRecvRL_State_WAITMSG) {
	if (AL_IRecvTest(&tmp) == MPI_SUCCESS) {
	  tmp->State = AL_IRecvRL_State_WAITMEM;
	}
      }
    } else
      alldone = AL_True;
  }
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_DATA)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_IRecvTreatPending : Processed %d entries\n", __AL_rank, __AL_log++, inc); _AL_ff;
#endif
#endif
  return(alldone);
}

/********************/

void
_AL_IRecvAddPending(AL_IRecvRL **rlist)
{ /* _AL_IRecvAddPending */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] _AL_IRecvAddPending [rlist : %X]\n", __AL_rank, __AL_log++, *rlist); _AL_ff;
#endif
#endif  
  
  (*rlist)->State = AL_IRecvRL_State_PENDING;
  AL_PendingIRecvRL_AddRlist(&AL_PendingIRecvRLHandler, *rlist);
  
  _AL_IRecvTreatPending(rlist);
}

/********************/

int
AL_IRecvTest(AL_IRecvRL **rlist)
{ /* AL_IRecvTest */

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvTest [rlist : %X] [State : %d]\n", __AL_rank, __AL_log++, *rlist, (*rlist)->State); _AL_ff;
#endif
#endif

  /* Pre-cond : we have a rlist to rely on */
  if (*rlist == NULL)
    return (_AL_ERROR_);

  /* First, let's treat the pending entries */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
    fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvTest : Treat Pending Entries\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  _AL_IRecvTreatPending(rlist);

  if ((*rlist)->State < AL_IRecvRL_State_WAITMSG) {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvTest : Exit [rlist : %X] [State : %d]\n", __AL_rank, __AL_log++, *rlist, (*rlist)->State); _AL_ff;
#endif
#endif
    return (_AL_ERROR_);
  }

  /* Then let's test all the remaining messages */
  if ((*rlist)->State == AL_IRecvRL_State_WAITMSG) {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvTest : Test all messages\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
    if (AL_IIRL_Testall((*rlist)->IIRL) != AL_True)
      return (_AL_ERROR_);

    (*rlist)->State = AL_IRecvRL_State_WAITMEM;
  }

  /* Everything went smoothly (or we are already past previous stages) */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvTest : return (MPI_SUCCESS)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  return (MPI_SUCCESS);
}

/********************/

int
AL_IRecvWait(AL_IRecvRL **rlist)
{ /* AL_IRecvWait */

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvWait [rlist : %X] [State : %d]\n", __AL_rank, __AL_log++, *rlist, (*rlist)->State); _AL_ff;
#endif
#endif

  /* Pre-cond : we have a rlist to rely on */
  if (*rlist == NULL)
    return (_AL_ERROR_);

  /* First, let's treat the pending entries */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
    fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvWait : Treat Pending Entries\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  _AL_IRecvTreatPending(rlist);
  while ((*rlist)->State < AL_IRecvRL_State_WAITMSG) {
    _AL_IRecvTreatPending(rlist);
  }

  /* Then let's wait on all the remaining messages */
  if ((*rlist)->State == AL_IRecvRL_State_WAITMSG) {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvWait : Wait on all messages\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
    if (AL_IIRL_Waitall((*rlist)->IIRL) != AL_True)
      return (_AL_ERROR_);

    (*rlist)->State = AL_IRecvRL_State_WAITMEM;
  }

  /* Everything went smoothly (or we are already past previous stages) */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvWait : return (MPI_SUCCESS)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  return (MPI_SUCCESS);
}

/********************/

int
AL_IRecvComplete(AL_IRecvRL **rlist)
{ /* AL_IRecvComplete */
  int tmp;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvComplete [rlist : %X] [State : %d]\n", __AL_rank, __AL_log++, *rlist, (*rlist)->State); _AL_ff;
#endif
#endif

  /* Pre-cond : we have a rlist to rely on */
  if (*rlist == NULL)
    return (_AL_ERROR_);

  /* First, let's make sure we have finished */
  if ((*rlist)->State != AL_IRecvRL_State_WAITMEM) {
    tmp = AL_IRecvWait(rlist);
    
    if (tmp != MPI_SUCCESS)
      return (tmp);
  }

  /* Let's recreate elements in memory */
  if ((*rlist)->State == AL_IRecvRL_State_WAITMEM) {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvComplete : Recreation of elements in memory\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

    _AL_IRecv_MemCreate(rlist);

    (*rlist)->State = AL_IRecvRL_State_WAITLINKS;
  }

  /* Now, let's recreate all the links */
  if ((*rlist)->State == AL_IRecvRL_State_WAITLINKS) {
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
    fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvComplete : Links reconstruction\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

    _AL_IRecvLink(rlist);

    (*rlist)->State = AL_IRecvRL_State_DONE;
  }

  /* First, let's make sure we have finished */
  if ((*rlist)->State != AL_IRecvRL_State_DONE)
    return (_AL_ERROR_);

  /* Result is Intial Object with recreated links */
  *((*rlist)->buf) = AL_SArray_GetValue((*rlist)->ALint->AL_Address,
					 (*rlist)->InitMsg->InitialDatatype, 0);
  *((*rlist)->datatype) = (*rlist)->InitMsg->InitialDatatype;

  /* Then let's give back all the used blocks */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvComplete : Give back all used blocks\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  AL_IRecvRL_GiveBlockBuffer(*rlist);

  /* Finally, let's empty the Pointer */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvComplete : Cleaning memory\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  _AL_RecvErrorFree((*rlist)->ALint);
  AL_PendingIRecvRL_GiveBack(&AL_PendingIRecvRLHandler, *rlist);
  AL_IRecvRL_Delete(*rlist);
  *rlist = NULL;

  /* Everything went smoothly */
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_IRecvComplete : return (MPI_SUCCESS)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif
  return (MPI_SUCCESS);
}

/********************/

int
AL_IRecv(void** buf, int* datatype, int source, int tag, MPI_Comm comm,
	 AL_IRecvRL** rlist, AL_ReqList** rqlist)
{ /* AL_IRecv */

  int i;

  /* AutoLink internals */
  AL_Internals  *ALint;

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FNAME)
  fprintf(__AL_file, "[%.4d/%.8d] AL_IRecv : [Source Rank : %d] [Tag : %d]\n", __AL_rank, __AL_log++, source, tag); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_IRecv Start\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_IRecv : Create AL internals needed\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#endif

  /* Create AL internals needed */
  ALint = (AL_Internals *) malloc(sizeof(AL_Internals));

  _AL_IRecvMalloc(ALint);
  
  for(i = 0; i < _AL_TpUsed; i++) {
    ALint->AL_DataCount[i] = 0;
  }

  /* Create a request list if required */
  if ((rqlist != NULL) && (*rqlist == NULL))
    *rqlist = (AL_ReqList *) AL_ReqList_Create();

  *rlist = AL_IRecvRL_Create(buf, datatype, source, tag, comm,  ALint, rqlist);
  
  _AL_IRecvAddPending(rlist);

#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_IRecv [rlist : %X] [State : %d]\n", __AL_rank, __AL_log++, *rlist, (*rlist)->State); _AL_ff;
#endif
#endif
  
#ifdef _AL_DEBUG_
#if (_AL_DBG_LVL & _AL_DBG_FINT)
  fprintf(__AL_file, "[%.4d/%.8d] AL_IRecv : return (MPI_SUCCESS)\n", __AL_rank, __AL_log++); _AL_ff;
#endif
#if  (_AL_DBG_LVL & _AL_DBG_BENCH)
  __AL_endtime = MPI_Wtime();
  fprintf(__AL_file, "[%.4d/%.8d] [ALBM %f : %d] AL_IRecv Finish\n", __AL_rank, __AL_log++, __AL_endtime - __AL_starttime, __AL_PacketSize); _AL_ff;
#endif
#endif

  return (MPI_SUCCESS);
}
